# -*- coding: utf-8 -*-

"""
EconoMe - Connecting your QGIS-Project with EconoMe.

Copyright (C) 2025 Federal Office for the Environment FOEN, Switzerland

Authors
  - Kevin Helzel (WSL Institute for Snow and Avalanche Research SLF, Davos, Switzerland)
  - Tobia Lezuo (Geotest AG, Davos, Switzerland)

Email
  - Plugin & Code: kevin.helzel@slf.ch
  - EconoMe project leader: linda.ettlin@slf.ch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.

Generated by QGIS Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
"""

# QGIS Main Imports
from qgis.PyQt.QtCore import Qt, QCoreApplication, QSettings, QTranslator, QVariant, QTimer
from qgis.PyQt.QtGui import QIcon, QPixmap, QColor, QCursor
from qgis.PyQt.QtWidgets import QAction, QDialog, QVBoxLayout, QHBoxLayout, QLabel, QPushButton, QComboBox, QSizePolicy, QSpacerItem
from qgis.core import QgsProject, QgsFeatureRequest, QgsMapLayerProxyModel, QgsVectorLayer, QgsField, QgsRendererRange, QgsGraduatedSymbolRenderer, QgsSymbol, NULL
# QGIS Processing Imports
import processing
from processing.tools import dataobjects

# Custom imports
import os.path
from datetime import datetime
import tempfile
import json
import requests
from requests.auth import HTTPBasicAuth
from collections import Counter
import copy

# Initialize Qt resources from file resources.py
from .resources import *

# Import dialogs
from .econo_me_dialog import EconoMeDialog
from .econo_me_dialog_status import StatusDialog
from .econo_me_dialog_asset_handling import DamagePotentialMismatchResolveDialog
from .econo_me_dialog_attribute_profile_building import EconoMeDialogAttributeProfileBuilding
from .econo_me_dialog_attribute_profile_road import EconoMeDialogAttributeProfileRoad
from .econo_me_dialog_attribute_profile_railway import EconoMeDialogAttributeProfileRailway
from .econo_me_dialog_attribute_profile_cable_transport import EconoMeDialogAttributeProfileCableTransport
from .econo_me_dialog_attribute_profile_line_pipe import EconoMeDialogAttributeProfileLinePipe
from .econo_me_dialog_attribute_profile_other_asset import EconoMeDialogAttributeProfileOtherAsset

class AreYouSure(QDialog):
    """Custom dialog for plugin status information."""

    def __init__(self, message, parent=None):

        # Initialize class
        super(AreYouSure, self).__init__(parent)

        # Dialog header
        self.setWindowTitle("+++ PLUGIN STATUS +++")
        self.setWindowIcon(QIcon(os.path.join(os.path.dirname(__file__), "icon.png")))

        # Main layout
        main_layout = QVBoxLayout()
        # Add a label
        label_layout = QHBoxLayout()
        label = QLabel(message)
        label.setAlignment(Qt.AlignHCenter)
        label.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        label_layout.addSpacerItem(QSpacerItem(10, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))
        label_layout.addWidget(label)
        label_layout.addSpacerItem(QSpacerItem(10, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))
        # Add an "DELETE"  and "CANCEL" button
        button_layout = QHBoxLayout()
        delete_button = QPushButton("LÖSCHEN")
        cancel_button = QPushButton("ABBRECHEN")
        delete_button.setCursor(QCursor(Qt.PointingHandCursor))
        cancel_button.setCursor(QCursor(Qt.PointingHandCursor))
        delete_button.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        cancel_button.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        button_layout.addSpacerItem(QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))
        button_layout.addWidget(delete_button)
        button_layout.addWidget(cancel_button)
        button_layout.addSpacerItem(QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))
        
        delete_button.clicked.connect(self.accept)
        cancel_button.clicked.connect(self.reject)
        
        # Add horizontal layouts to main layout
        main_layout.addLayout(label_layout)
        main_layout.addLayout(button_layout)

        # Set main layout
        self.setLayout(main_layout)

class EconoMe:
    """QGIS Plugin Implementation."""

    def __init__(self, iface):

        # Save reference to the QGIS interface
        self.iface = iface
        # Initialize plugin directory
        self.plugin_dir = os.path.dirname(__file__)
        
        # Own variables
        self.ECONOME_USER_INFO = None
        self.ECONOME_USER_INFO_FILE_PATH = None
        self.ECONOME_API_PROJECT_BASE_PATH = "https://api.econome.ch/V2/project"
        self.ECONOME_API_PROFILE_BASE_PATH = "https://api.econome.ch/V2/profile"
        self.ECONOME_PROJECT_NUMBER = None
        self.ECONOME_PROJECT_LIST = None
        self.ECONOME_USER_NAME = None
        self.ECONOME_PASSWORD = None
        self.LOGIN_ERROR_MESSAGE = None
        self.USER_INFO_VALIDATED = False
        self.TIMER = None
        
        self.DAMAGE_POTENTIAL_LAYER_LIST = None
        self.DAMAGE_POTENTIAL_ASSETS_HANDLING_OPTION = ""
        self.DAMAGE_POTENTIAL_LAYER_STATUS = {
            "BUILDING": 0,
            "ROAD": 0,
            "RAILWAY": 0,
            "CABLE_TRANSPORT": 0,
            "LINE_PIPE": 0,
            "OTHER_ASSET": 0
        }
        self.DAMAGEPOTENTIAL_REQUIRED_FIELDS = {
            "building": ["AssetIdentID", "AssetTypeID"],
            "road": ["AssetIdentID", "AssetTypeID", "DailyTraffic", "Speed"],
            "railway": ["AssetIdentID", "AssetTypeID", "DailyTraffic", "TrainLength"],
            "cable_transport": ["AssetIdentID", "AssetTypeID", "PassengerVolume", "Speed"],
            "line_pipe": ["AssetIdentID", "AssetTypeID"],
            "other_asset": ["AssetIdentID", "AssetTypeID"]
        }
        self.ASSET_CATEGORIES_BY_DAMAGEPOTENTIAL_TYPE = {
            "BUILDING": ['Gebäude', 'Sonderobjekte'],
            "ROAD": ['Strassenverkehr'],
            "RAILWAY": ['Schienenverkehr'],
            "CABLE_TRANSPORT": ['Mechanische Aufstiegshilfe'],
            "LINE_PIPE": ['Leitungen'],
            "OTHER_ASSET": ['Landwirtschaft, Wald und Grünanlagen', 'Infrastruktur']
        }
        self.ASSET_TYPEID_NAME_BY_CATEGORYNAME = {
            'Gebäude': {
                'Bahnhof': 14,
                'Einkaufszentrum': 9,
                'Frei wählbares Gebäude': 64,
                'Garage (Parkeinheit)': 4,
                'Hotel': 5,
                'Industrie-/ Gewerbegebäude': 6,
                'Keller (nur für Wasserprozesse)': 91,
                'Kirche': 12,
                'Öffentliches Gebäude': 18,
                'Öffentlicher Parkplatz': 88,
                'Schule / Kindergarten': 10,
                'Schuppen / Remise': 3,
                'Spital': 11,
                'Sportanlage (Gebäude)': 15,
                'Stationsgebäude mechanische Aufstiegshilfe': 63,
                'Viehstall': 2,
                'Wohneinheit Einfamilienhaus': 1,
                'Wohneinheit Mehrfamilienhaus': 87
            },
            'Sonderobjekte': {
                'ARA': 19,
                'Campingplatz': 21,
                'Zeltplatz': 89,
                'Deponie': 20,
                'Frei wählbares Sonderobjekt': 66,
                'Kraftwerk': 8,
                'Sendeanlage': 65,
                'Wasserreservoir': 22
            },
            'Strassenverkehr': {
                'Brücke Autobahn': 34,
                'Brücke Einzelfahrzeuge': 31,
                'Brücke Kantonsstrasse': 33,
                'Brücke Gemeindestrasse': 32,
                'Asphaltweg': 56,
                'Kiesweg': 55,
                'Frei wählbares Strassenverkehrsobjekt': 67,
                'Gemeindestrasse': 26,
                'Kantonsstrasse': 25,
                'Nationalstrasse': 24
            },
            'Leitungen': {
                'Abwasser unter Terrain': 60,
                'Frei wählbares Leitungsobjekt': 82,
                'Freileitung': 38,
                'Gas auf Terrain': 61,
                'Gas unter Terrain': 80,
                'Hochspannungsleitung <= 150 kV': 57,
                'Hochspannungsleitung > 150 kV': 58,
                'Strom unter Terrain': 37,
                'Telekommunikation auf Terrain': 62,
                'Telekommunikation unter Terrain': 81,
                'Wasser auf Terrain': 59,
                'Wasser unter Terrain': 79
            },
            'Mechanische Aufstiegshilfe': {
                'Frei wählbares Objekt (mech. Aufstiegshilfe)': 78,
                'Gondelbahn': 75,
                'Luftseilbahn': 76,
                'Sessellift': 74,
                'Skilift': 73,
                'Standseilbahn': 77
            },
            'Landwirtschaft, Wald und Grünanlagen': {
                'extensive Flächen': 43,
                'Frei wählbares Objekt (Landwirtschaft, Grünflächen, Wald)': 86,
                'Friedhof': 84,
                'Gemüsekultur': 47,
                'Golfanlage': 83,
                'intensive Flächen': 41,
                'Nutzwald': 45,
                'Obstplantage': 48,
                'Parkanlage': 42,
                'Rebberg': 46,
                'Schutzwald': 44,
                'Sportaussenanlage': 85
            },
            'Schienenverkehr': {
                'Brücke Einspur': 68,
                'Brücke Mehrspur': 69,
                'Einspur': 28,
                'Mehrspur': 29,
                'Sonderobjekt Brücke Schienenverkehr': 90
            },
            'Infrastruktur': {
                'Durchtrennung Kabelanlage (pro Schadenstelle)': 71,
                'Bahninfrastruktur': 72,
                'Schaltposten': 70,
                'Schutzbauwerk': 54
            }
        }
        self.QVARIANT_TYPES_DICT = {
            "integer": [QVariant.Int, QVariant.LongLong], 
            "float": [QVariant.Double], 
            "string": [QVariant.String], 
            "boolean": [QVariant.Bool]
        } 
        self.AFFECTED_ASSET_SPECIAL_PARAMETERS = {
            "BUILDING": [("SOL_ID", "integer"), ("p_rA", "number"), ("p_rA_Comment", "string"), ("PresenceValue", "numZeroToOne"), ("PresenceComment", "string"), ("ProtectionValue", "numZeroToOne"), ("ProtectionComment", "string")],
            "ROAD": [("SOL_ID", "integer"), ("p_rA", "number"), ("p_rA_Comment", "string"), ("p_vSp", "numZeroToOne"), ("p_vSp_Comment", "string"), ("AlarmSystem", "boolean")],
            "RAILWAY": [("SOL_ID", "integer"), ("p_rA", "number"), ("p_rA_Comment", "string"), ("p_vSp", "numZeroToOne"), ("p_vSp_Comment", "string"), ("AlarmSystem", "boolean")],
            "CABLE_TRANSPORT": [("SOL_ID", "integer"), ("p_rA", "number"), ("p_rA_Comment", "string")],
            "LINE_PIPE": [("SOL_ID", "integer"), ("p_rA", "number"), ("p_rA_Comment", "string")],
            "OTHER_ASSET": [("SOL_ID", "integer"), ("p_rA", "number"), ("p_rA_Comment", "string")]
        }
        self.RISK_OUTPUT_VARIABLES = {
            "BUILDING": ["R_Sw", "R_P(CHF)", "R", "r_indiv"],
            "ROAD": ["R_Sw", "R_P(CHF)", "R", "r_indiv", "R_Dir"],
            "RAILWAY": ["R_Sw", "R_P(CHF)", "R", "r_indiv", "R_Dir", "R_Anp"],
            "CABLE_TRANSPORT": ["R_Sw", "R_P(CHF)", "R", "r_indiv"],
            "LINE_PIPE": ["R_Sw"],
            "OTHER_ASSET": ["R_Sw"]
        }

        # Initialize locale
        locale = QSettings().value('locale/userLocale')[0:2]
        locale_path = os.path.join(
            self.plugin_dir,
            'i18n',
            'EconoMe_{}.qm'.format(locale))

        if os.path.exists(locale_path):
            self.translator = QTranslator()
            self.translator.load(locale_path)
            QCoreApplication.installTranslator(self.translator)

        # Declare instance attributes
        self.actions = []
        self.menu = self.tr(u'&EconoMe')

        # Check if plugin was started the first time in current QGIS session
        # Must be set in initGui() to survive plugin reloads
        self.first_start = None

    # noinspection PyMethodMayBeStatic
    def tr(self, message):
        """Get the translation for a string using Qt translation API."""
        # noinspection PyTypeChecker,PyArgumentList,PyCallByClass
        return QCoreApplication.translate('EconoMe', message)

    def add_action(
            self,
            icon_path,
            text,
            callback,
            enabled_flag=True,
            add_to_menu=True,
            add_to_toolbar=True,
            status_tip=None,
            whats_this=None,
            parent=None):

        icon = QIcon(icon_path)
        action = QAction(icon, text, parent)
        action.triggered.connect(callback)
        action.setEnabled(enabled_flag)

        if status_tip is not None:
            action.setStatusTip(status_tip)

        if whats_this is not None:
            action.setWhatsThis(whats_this)

        action.setToolTip("EconoMe")
        action.setWhatsThis("EconoMe QGIS Plugin")

        if add_to_toolbar:
            # Adds plugin icon to Plugins toolbar
            self.iface.addToolBarIcon(action)

        if add_to_menu:
            self.iface.addPluginToMenu(
                self.menu,
                action)

        self.actions.append(action)

        return action

    def initGui(self):
        """Create the menu entries and toolbar icons inside the QGIS GUI."""

        icon_path = os.path.join(os.path.dirname(__file__), "icon.png")
        self.add_action(
            icon_path,
            text=self.tr(u'EconoMe'),
            callback=self.run,
            parent=self.iface.mainWindow())

        # will be set False in run()
        self.first_start = True

    def unload(self):
        """Removes the plugin menu item and icon from QGIS GUI."""
        for action in self.actions:
            self.iface.removePluginMenu(
                self.tr(u'&EconoMe'),
                action)
            self.iface.removeToolBarIcon(action)

    def get_econome_user_info(self):
        """Reads .txt file containing EconoMe user info (project numbers and API keys)"""

        # CHECK: Existing user info file
        prefix = "econome_user_info_"
        files_in_directory = os.listdir(self.plugin_dir)
        matching_files = [file for file in files_in_directory if file.startswith(prefix) and file.endswith(".txt")]
        # If there is a file, read and return EconoMe user infos
        if matching_files:
            # Get file path
            self.ECONOME_USER_INFO_FILE_PATH = os.path.join(self.plugin_dir, matching_files[0])
            # Read out data
            file = open(self.ECONOME_USER_INFO_FILE_PATH, "r")
            info_ID = file.readline().strip()
            info_ID_list = file.readline().strip()
            info_KeyA = file.readline().strip()
            info_KeyB = file.readline().strip()
            file.close()
            # Return EconoMe user infos
            return {"ECONOME_PROJECT_NUMBER": info_ID, "ECONOME_PROJECT_LIST": info_ID_list.split(', '), "ECONOME_USER_NAME": info_KeyA, "ECONOME_PASSWORD": info_KeyB}
        else:
            # Return an empty dictionary (if no file is found)
            self.ECONOME_USER_INFO_FILE_PATH = None
            return None

    def validate_econome_user_info(self):
        """Checks if API keys were given and are valid."""

        # Get user name and password from lineEdit
        ECONOME_USER_NAME = self.main_dialog.LINE_EDIT_ECONOME_USER_NAME.text().strip()
        ECONOME_PASSWORD = self.main_dialog.LINE_EDIT_ECONOME_PASSWORD.text().strip()

        # CHECK: User name and password filled
        if ECONOME_USER_NAME and ECONOME_PASSWORD:
            # Make test API request
            make_test_request = requests.get(self.ECONOME_API_PROJECT_BASE_PATH, auth=HTTPBasicAuth(ECONOME_USER_NAME, ECONOME_PASSWORD))
            TEST_REQUEST = json.loads(make_test_request.text)
            # Test API request was successful
            if 'ProjectList' in TEST_REQUEST:
                # Set EconoMe user info
                self.ECONOME_USER_NAME = ECONOME_USER_NAME
                self.ECONOME_PASSWORD = ECONOME_PASSWORD
                self.ECONOME_PROJECT_LIST = [(proj['ProjectIdentID'], proj['ProjectTitle']) for proj in TEST_REQUEST['ProjectList']]
                self.ECONOME_PROJECT_NUMBER = self.ECONOME_PROJECT_LIST[0]
                # Fill dropdown with ProjectIdentIDs
                self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.clear()
                for project_info in self.ECONOME_PROJECT_LIST:
                    if project_info[0] not in [self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.itemText(i).split(', ') for i in range(self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.count())]:
                        self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.addItem(f"{project_info[1]}, {project_info[0]}")
                # Set first project as current project
                self.ECONOME_PROJECT_NUMBER = self.ECONOME_PROJECT_LIST[0][0]
                self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.setCurrentText(self.ECONOME_PROJECT_NUMBER)
                # Save project infos in plugin folder (if they were not existent)
                if not self.get_econome_user_info():
                    # Create a temporary file in the plugin folder
                    self.ECONOME_USER_INFO_FILE_PATH = tempfile.NamedTemporaryFile(dir=self.plugin_dir, prefix="econome_user_info_", suffix=".txt", delete=False)
                    # Write project infos to a temporary text file which persists until QGIS is closed
                    with open(self.ECONOME_USER_INFO_FILE_PATH.name, "w") as file:
                        file.write(f"{self.ECONOME_PROJECT_NUMBER}\n")
                        file.write(f"{', '.join([item[0] for item in self.ECONOME_PROJECT_LIST])}\n")
                        file.write(f"{ECONOME_USER_NAME}\n")
                        file.write(f"{ECONOME_PASSWORD}\n")
                        file.flush()
                # If existing user info, check for differences and overwrite
                else:
                    existing_user_info = self.get_econome_user_info()
                    if existing_user_info['ECONOME_PROJECT_NUMBER'] != self.ECONOME_PROJECT_NUMBER or existing_user_info['ECONOME_PROJECT_LIST'] != self.ECONOME_PROJECT_LIST or existing_user_info['ECONOME_USER_NAME'] != self.ECONOME_USER_NAME or existing_user_info['ECONOME_PASSWORD'] != self.ECONOME_PASSWORD:
                        # Overwrite project infos to a temporary text file which persists until QGIS is closed
                        with open(os.path.join(self.plugin_dir, self.ECONOME_USER_INFO_FILE_PATH), "w") as file:
                            file.write(f"{self.ECONOME_PROJECT_NUMBER}\n")
                            file.write(f"{', '.join([item[0] for item in self.ECONOME_PROJECT_LIST])}\n")
                            file.write(f"{ECONOME_USER_NAME}\n")
                            file.write(f"{ECONOME_PASSWORD}\n")
                            file.flush()
                # Set new label text
                self.main_dialog.LABEL_ECONOME_PROJECT_NUMBER.setText("Projekt(e)")
                # Color the login button green
                self.main_dialog.BUTTON_ECONOME_PROJECT_CONNECT.setStyleSheet("background-color: #27ae60;")
                # Black color for combo box items
                self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.setStyleSheet("QComboBox:editable { color: black; }")
                # EconoMe user info -> validated
                self.USER_INFO_VALIDATED = True
                # Get project status
                self.get_project_status()
            # Test API request was not successful
            else:
                # EconoMe user info -> not validated
                self.USER_INFO_VALIDATED = False
                # Remove green color from login button
                self.main_dialog.BUTTON_ECONOME_PROJECT_CONNECT.setStyleSheet("background-color: none;")
                # Receive error messages and display them in the combo box
                if "Errorcode" in TEST_REQUEST:
                    self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.clear()
                    self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.addItem(f"{TEST_REQUEST['Errormessage']}")
                    self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.setStyleSheet("QComboBox:editable { color: red; }")
                else:
                    self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.clear()
                    self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.addItem(f"{TEST_REQUEST['Message']}")
                    self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.setStyleSheet("QComboBox:editable { color: red; }")
                # Get project status
                self.get_project_status()
        # User name and/or password were not filled
        else:
            # EconoMe user info -> not validated
            self.USER_INFO_VALIDATED = False
            # Get project status
            self.get_project_status()

    def onChange_project_selection(self):
        """Switches project number and updates EconoMe user info."""

        # CHECK: EconoMe user info validated
        if self.USER_INFO_VALIDATED:
            # CHECK: Project selected
            if self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.currentText():
                # CHECK: Project selection changed
                if self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.currentText().split(', ')[1] != self.ECONOME_PROJECT_NUMBER:
                    # Set new project number
                    self.ECONOME_PROJECT_NUMBER = self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.currentText().split(', ')[1]
                    # Overwrite project infos in the temporary text file which persists until QGIS is closed
                    if isinstance(self.ECONOME_USER_INFO_FILE_PATH, str):
                        with open(self.ECONOME_USER_INFO_FILE_PATH, "w") as file:
                            file.write(f"{self.ECONOME_PROJECT_NUMBER}\n")
                            file.write(f"{', '.join([item[0] for item in self.ECONOME_PROJECT_LIST])}\n")
                            file.write(f"{self.ECONOME_USER_NAME}\n")
                            file.write(f"{self.ECONOME_PASSWORD}\n")
                            file.flush()
                    else:
                        with open(self.ECONOME_USER_INFO_FILE_PATH.name, "w") as file:
                            file.write(f"{self.ECONOME_PROJECT_NUMBER}\n")
                            file.write(f"{', '.join([item[0] for item in self.ECONOME_PROJECT_LIST])}\n")
                            file.write(f"{self.ECONOME_USER_NAME}\n")
                            file.write(f"{self.ECONOME_PASSWORD}\n")
                            file.flush()
                    # Get project status
                    self.get_project_status()

    def onChange_econome_user_info(self):
        """Resets plugin UI when EconoMe user name or password is modified."""

        # CHECK: EconoMe user info validated
        if self.USER_INFO_VALIDATED:
            # EconoMe user info -> not validated
            self.USER_INFO_VALIDATED = False
            # Clear project combo box
            self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.clear()
            # Display messsage -> not connected
            self.main_dialog.LABEL_ECONOME_PROJECT_NUMBER.setText("Projektstatus ")
            self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.addItem("Nicht verbunden")
            # Remove green color from login button
            self.main_dialog.BUTTON_ECONOME_PROJECT_CONNECT.setStyleSheet("background-color: none;")
            # Get project status
            self.get_project_status()

    def get_project_status(self):
        """Retrieves EconoMe project status information and fills plugin UI."""

        # Create project status dictionary
        PROJECT_STATUS = {
            "HazardProcesses": {
                "Count": 0
            },
            "Variants": {
                "Count": 0
            },
            "Scenarios": {
                "Active": []
            }
        }

        # CHECK: EconoMe user info validated
        if self.USER_INFO_VALIDATED:
            # (1) Get hazard processes
            get_hazard_processes = requests.get(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + '/process', auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
            HAZARD_PROCESSES = json.loads(get_hazard_processes.text)
            if isinstance(HAZARD_PROCESSES, list):
                PROJECT_STATUS["HazardProcesses"]["Count"] = len(HAZARD_PROCESSES)
                self.main_dialog.LABEL_HAZARD_PROCESSES.setText(f"{len(HAZARD_PROCESSES)} GEFAHRENPROZESS(E)")
                hazard_processes_list = [(process["ProcessID"], process["ProcessType"], process["ProcessName"]) for process in HAZARD_PROCESSES]
                self.main_dialog.COMBO_BOX_HAZARD_PROCESS.clear()
                # Fill the drop-down with all hazard processes
                for hazard_process in hazard_processes_list:
                    if f"{hazard_process[0]}, {hazard_process[1]}, {hazard_process[2]}" not in [self.main_dialog.COMBO_BOX_HAZARD_PROCESS.itemText(i) for i in range(self.main_dialog.COMBO_BOX_HAZARD_PROCESS.count())]:
                        self.main_dialog.COMBO_BOX_HAZARD_PROCESS.addItem(f"{hazard_process[0]}, {hazard_process[1]}, {hazard_process[2]}")
                        self.main_dialog.COMBO_BOX_HAZARD_PROCESS.setCurrentText(f"{hazard_process[0]}, {hazard_process[1]}, {hazard_process[2]}")

                # Get selected hazard process
                selected_hazard_process = self.main_dialog.COMBO_BOX_HAZARD_PROCESS.currentText()
                selected_hazard_process_info = [hazard_process for hazard_process in HAZARD_PROCESSES if hazard_process["ProcessID"] == int(selected_hazard_process.split(",")[0])]
                if len(selected_hazard_process_info) != 0:
                    # Set annualities for the two free scenarios
                    self.main_dialog.LABEL_SCENARIO_FREE.setText("  " + str(selected_hazard_process_info[0]["AnnualityFree"]["Annuality"]) + '-jährlich')
                    self.main_dialog.LABEL_SCENARIO_EXTREME.setText(str(selected_hazard_process_info[0]["AnnualityExtreme"]["Annuality"]) + '-jährlich')

            # (2) Get variants
            get_measure_variants = requests.get(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + '/variant', auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
            MEASURE_VARIANTS = json.loads(get_measure_variants.text)
            if isinstance(MEASURE_VARIANTS, list):
                PROJECT_STATUS["Variants"]["Count"] = len(MEASURE_VARIANTS)
                self.main_dialog.LABEL_MEASURE_VARIANTS.setText(f"{PROJECT_STATUS['Variants']['Count']} MASSNAHMENVARIANTE(N)")
                measure_variants_list = [(variant["VariantID"], variant["VariantName"]) for variant in MEASURE_VARIANTS]
                self.main_dialog.COMBO_BOX_MEASURE_VARIANT.clear()
                measure_variants_list.insert(0, (0, "Vor Massnahme"))
                for measure_variant in measure_variants_list:
                    if f"{measure_variant[0]}, {measure_variant[1]}" not in [self.main_dialog.COMBO_BOX_MEASURE_VARIANT.itemText(i) for i in range(self.main_dialog.COMBO_BOX_MEASURE_VARIANT.count())]:
                        self.main_dialog.COMBO_BOX_MEASURE_VARIANT.addItem(f"{measure_variant[0]}, {measure_variant[1]}")
                        self.main_dialog.COMBO_BOX_MEASURE_VARIANT.setCurrentText(f"{measure_variant[0]}, {measure_variant[1]}")

            
            return PROJECT_STATUS

        # EconoMe user info not validated
        else:
            # Clear labels
            self.main_dialog.IMAGE_LABEL_DAM_POT_STATUS_BUILDING.clear()
            self.main_dialog.IMAGE_LABEL_DAM_POT_STATUS_ROAD.clear()
            self.main_dialog.IMAGE_LABEL_DAM_POT_STATUS_RAILWAY.clear()
            self.main_dialog.IMAGE_LABEL_DAM_POT_STATUS_CABLE_TRANSPORT.clear()
            self.main_dialog.IMAGE_LABEL_DAM_POT_STATUS_LINE_PIPE.clear()
            self.main_dialog.IMAGE_LABEL_DAM_POT_STATUS_OTHER_ASSET.clear()
            self.main_dialog.IMAGE_LABEL_SCENARIO_STATUS_FREE.clear()
            self.main_dialog.IMAGE_LABEL_SCENARIO_STATUS_30.clear()
            self.main_dialog.IMAGE_LABEL_SCENARIO_STATUS_100.clear()
            self.main_dialog.IMAGE_LABEL_SCENARIO_STATUS_300.clear()
            self.main_dialog.IMAGE_LABEL_SCENARIO_STATUS_EXTREME.clear()
            # Reset drop-downs
            self.main_dialog.LAYER_DAM_POT_BUILDING.setCurrentIndex(0)
            self.main_dialog.LAYER_DAM_POT_ROAD.setCurrentIndex(0)
            self.main_dialog.LAYER_DAM_POT_RAILWAY.setCurrentIndex(0)
            self.main_dialog.LAYER_DAM_POT_CABLE_TRANSPORT.setCurrentIndex(0)
            self.main_dialog.LAYER_DAM_POT_LINE_PIPE.setCurrentIndex(0)
            self.main_dialog.LAYER_DAM_POT_OTHER_ASSET.setCurrentIndex(0)
            self.main_dialog.COMBO_BOX_HAZARD_PROCESS.clear()
            self.main_dialog.COMBO_BOX_MEASURE_VARIANT.clear()
            self.main_dialog.LAYER_INTENSITY_SCENARIO_FREE.setCurrentIndex(0)
            self.main_dialog.LAYER_INTENSITY_SCENARIO_30.setCurrentIndex(0)
            self.main_dialog.LAYER_INTENSITY_SCENARIO_100.setCurrentIndex(0)
            self.main_dialog.LAYER_INTENSITY_SCENARIO_300.setCurrentIndex(0)
            self.main_dialog.LAYER_INTENSITY_SCENARIO_EXTREME.setCurrentIndex(0)
            # Reset drop-down texts
            self.main_dialog.LABEL_HAZARD_PROCESSES.setText("GEFAHRENPROZESS(E)")
            self.main_dialog.LABEL_MEASURE_VARIANTS.setText("MASSNAHMENVARIANTE(N)")
            # Reset custom scenario labels
            self.main_dialog.LABEL_SCENARIO_FREE.setText('  -jährlich')
            self.main_dialog.LABEL_SCENARIO_EXTREME.setText('-jährlich')
            # Return empty dictionary
            return {}

    def set_parameter_profiles(self, pp_type):
        """Handles parameter profile dialogs."""

        DAMAGE_POTENTIAL_LAYER = None
        dialog = None

        if pp_type == "building":

            DAMAGE_POTENTIAL_LAYER = self.main_dialog.LAYER_DAM_POT_BUILDING.currentLayer()
            if DAMAGE_POTENTIAL_LAYER:
                dialog = EconoMeDialogAttributeProfileBuilding()
                dialog.setWindowTitle("Attributspalten: Gebäude")

        elif pp_type == "road":

            DAMAGE_POTENTIAL_LAYER = self.main_dialog.LAYER_DAM_POT_ROAD.currentLayer()
            if DAMAGE_POTENTIAL_LAYER:
                dialog = EconoMeDialogAttributeProfileRoad()
                dialog.setWindowTitle("Attributspalten: Strassen")

        elif pp_type == "railway":

            DAMAGE_POTENTIAL_LAYER = self.main_dialog.LAYER_DAM_POT_RAILWAY.currentLayer()
            if DAMAGE_POTENTIAL_LAYER:
                dialog = EconoMeDialogAttributeProfileRailway()
                dialog.setWindowTitle("Attributspalten: Schiene")

        elif pp_type == "cable_transport":

            DAMAGE_POTENTIAL_LAYER = self.main_dialog.LAYER_DAM_POT_CABLE_TRANSPORT.currentLayer()
            if DAMAGE_POTENTIAL_LAYER:
                dialog = EconoMeDialogAttributeProfileCableTransport()
                dialog.setWindowTitle("Attributspalten: Mech. Aufstiegshilfe")

        elif pp_type == "line_pipe":

            DAMAGE_POTENTIAL_LAYER = self.main_dialog.LAYER_DAM_POT_LINE_PIPE.currentLayer()
            if DAMAGE_POTENTIAL_LAYER:
                dialog = EconoMeDialogAttributeProfileLinePipe()
                dialog.setWindowTitle("Attributspalten: Leitungen")
        
        elif pp_type == "other_asset":

            DAMAGE_POTENTIAL_LAYER = self.main_dialog.LAYER_DAM_POT_OTHER_ASSET.currentLayer()
            if DAMAGE_POTENTIAL_LAYER:
                dialog = EconoMeDialogAttributeProfileOtherAsset()
                dialog.setWindowTitle("Attributspalten: Andere")

        if dialog:
            # find all labels of this pp_type
            labels = dialog.findChildren(QLabel)
            
            # find all fields in the layer
            fields = DAMAGE_POTENTIAL_LAYER.fields()
            integer_fields = [field for field in fields if field.type() in self.QVARIANT_TYPES_DICT["integer"]]
            number_fields = [field for field in fields if field.type() in self.QVARIANT_TYPES_DICT["float"]]
            string_fields = [field for field in fields if field.type() in self.QVARIANT_TYPES_DICT["string"]]
            bool_fields = [field for field in fields if field.type() in self.QVARIANT_TYPES_DICT["boolean"]]
            
            # Retrieve parameter profile
            get_parameter_profile = requests.get(self.ECONOME_API_PROFILE_BASE_PATH + f"/plugin-{pp_type}", auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
            PARAMETER_PROFILE = json.loads(get_parameter_profile.text)
            PARAMETER_PROFILE['Empty'] = False
            
            # if there is no parameter profile, create an empty one to have the type information
            if 'Code' in PARAMETER_PROFILE:
                empty_parameter_profile = {"ParameterList": [{"API": label.objectName().split('_')[1], "User": ""} for label in labels]}
                put_empty_parameter_profile = requests.post(self.ECONOME_API_PROFILE_BASE_PATH + f"/plugin-{pp_type}",
                                            json.dumps(empty_parameter_profile), auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                PARAMETER_PROFILE = json.loads(put_empty_parameter_profile.text)
                PARAMETER_PROFILE['Empty'] = True
            
            # Fill damage potential layer field names in combo boxes and select a parameter if set in the parameter profile
            for label in labels:
                attribute = label.objectName().split('_')[1]
                matching_parameter = [item for item in PARAMETER_PROFILE['ParameterList'] if item["API"] == attribute][0]
                parameter_type = matching_parameter['Type']
                label.setText(f"{label.text()} [{parameter_type}]")
                allowed_fields = integer_fields if parameter_type == "integer" else number_fields if parameter_type == "float" else string_fields if parameter_type == "string" else bool_fields if parameter_type == "boolean" else []
                matching_comboBox = dialog.findChild(QComboBox, f"COMBO_BOX_{attribute}")
                matching_comboBox.addItem("")
                matching_comboBox.addItems([field.name() for field in allowed_fields])
                if not PARAMETER_PROFILE['Empty']:
                    if f"{matching_parameter['User'].strip()}" in [field.name() for field in fields]:
                        index = matching_comboBox.findText(f"{matching_parameter['User'].strip()}")
                        matching_comboBox.setCurrentIndex(index)

            result = dialog.exec_()
            if result:
                # CREATE: Updated parameter profile dictionary
                updated_parameter_profile = {"ParameterList": []}
                # Loop over all user input combo boxes
                for comboBox in dialog.findChildren(QComboBox):
                    if not PARAMETER_PROFILE['Empty']:
                        existing_parameter_pair = [item for item in PARAMETER_PROFILE['ParameterList'] if item["API"] == comboBox.objectName().split("_")[-1]][0]
                        if existing_parameter_pair["User"] != comboBox.currentText():
                            updated_parameter_profile["ParameterList"].append({"API": existing_parameter_pair["API"], "User": comboBox.currentText()})
                    else:
                        parameter_name = comboBox.objectName().split("_")[-1]
                        updated_parameter_profile["ParameterList"].append({"API": parameter_name, "User": comboBox.currentText()})

                if len(updated_parameter_profile["ParameterList"]) != 0:
                    # create the pixmap
                    path_to_checkmark = os.path.join(self.plugin_dir, "general-icons", "checkmark.png")
                    pixmap_checkmark = QPixmap(path_to_checkmark)
                    self.scaled_pixmap_checkmark = pixmap_checkmark.scaled(
                        self.main_dialog.IMAGE_LABEL_SCENARIO_STATUS_FREE.size(),
                        Qt.KeepAspectRatio,
                        Qt.SmoothTransformation)
                    path_to_warning = os.path.join(self.plugin_dir, "general-icons", "warning.png")
                    pixmap_warning = QPixmap(path_to_warning)
                    self.scaled_pixmap_warning = pixmap_warning.scaled(
                        self.main_dialog.IMAGE_LABEL_SCENARIO_STATUS_FREE.size(),
                        Qt.KeepAspectRatio,
                        Qt.SmoothTransformation)
                    
                    if not PARAMETER_PROFILE['Empty']:
                        put_new_parameter_profile = requests.put(self.ECONOME_API_PROFILE_BASE_PATH + f"/plugin-{pp_type}",
                                                                 json.dumps(updated_parameter_profile), auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                        # USER INFO
                        if put_new_parameter_profile.status_code == 200:
                            StatusDialog(status_icon = self.scaled_pixmap_checkmark, message_title="Erfolg", message_text=f"\nAttributspalten für {dialog.windowTitle().split(': ')[-1]} erfolgreich aktualisiert!\n").exec_()
                        else:
                            StatusDialog(status_icon = self.scaled_pixmap_warning, message_title="Fehler", message_text=f"\n{put_new_parameter_profile.json()}\n").exec_()
                    else:
                        put_new_parameter_profile = requests.put(self.ECONOME_API_PROFILE_BASE_PATH + f"/plugin-{pp_type}",
                                                                   json.dumps(updated_parameter_profile), auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                        # USER INFO
                        if put_new_parameter_profile.status_code == 200:
                            StatusDialog(status_icon = self.scaled_pixmap_checkmark, message_title="Erfolg", message_text=f"\nAttributspalten für {dialog.windowTitle().split(': ')[-1]} erfolgreich angelegt!\n").exec_()
                        else:
                            StatusDialog(status_icon = self.scaled_pixmap_warning, message_title="Fehler", message_text=f"\n{put_new_parameter_profile.json()}\n").exec_()

    def compare_damage_potential(self, layer_assets, uploaded_damage_potential, valid_parameter_pairs):
        """This method compares the damage potential from a GIS layer with stored data in EconoMe."""
        # Check for existing uploaded damage potential
        if len(uploaded_damage_potential) != 0:
            # Create two "AssetIdentID" lists
            uploaded_AssetIdentIDs = [asset["AssetIdentID"] for asset in uploaded_damage_potential]
            layer_AssetIdentIDs = [str(asset["AssetIdentID"]) for asset in layer_assets["AssetList"]]
            # Create empty attribute mismatch "AssetIdentID" list
            mismatch_dict = {}
            # CASE: All uploaded assets are present in the damage potential layer
            if set(uploaded_AssetIdentIDs).issubset(layer_AssetIdentIDs):
                # Check for attribute differences
                for layer_asset in layer_assets["AssetList"]:
                    asset_missmatch_dict = {}
                    corresponding_uploaded_asset = [asset for asset in uploaded_damage_potential if asset["AssetIdentID"] == str(layer_asset["AssetIdentID"])]
                    if len(corresponding_uploaded_asset) > 0:
                        for pp in valid_parameter_pairs:
                            if pp["API"] in layer_asset.keys():
                                if layer_asset[pp["API"]] != corresponding_uploaded_asset[0][pp["API"]]:
                                    asset_missmatch_dict.update({pp["API"]: {"Layer": layer_asset[pp["API"]], "API": corresponding_uploaded_asset[0][pp["API"]]}})
                    else:
                        asset_missmatch_dict = {"Missing":"Fehlt in EconoMe."}
                    if asset_missmatch_dict != {}:
                        mismatch_dict.update({layer_asset["AssetIdentID"]: asset_missmatch_dict})
                if len(mismatch_dict.keys()) == 0:
                    return "COMPLETE MATCH", mismatch_dict
                else:
                    handling_dialog = DamagePotentialMismatchResolveDialog(mode = 'MISMATCH', mismatch = mismatch_dict, main_dialog = self.main_dialog)
                    if handling_dialog.exec_() == QDialog.Accepted:
                        return handling_dialog.selected_option, mismatch_dict
                    else:
                        return "CANCEL", {}
            # CASE: not all uploaded assets are present in the damage potential layer
            elif not set(layer_AssetIdentIDs).issubset(uploaded_AssetIdentIDs) or len(set(uploaded_AssetIdentIDs)) > len(layer_AssetIdentIDs):
                mismatch_dict = {AssetIdentID : {"Missing":"Fehlt im Layer."} for AssetIdentID in list(set(uploaded_AssetIdentIDs) - set(layer_AssetIdentIDs))}
                handling_dialog = DamagePotentialMismatchResolveDialog(mode = 'MISSING', mismatch = mismatch_dict, main_dialog = self.main_dialog)
                if handling_dialog.exec_() == QDialog.Accepted:
                    return handling_dialog.selected_option, mismatch_dict
                else:
                    return "CANCEL", {}
        else:
            # CASE ONLY POST: No uploaded damage potential existing
            return "POST LAYER ASSETS", {}

    def synchronize_damage_potential(self, damage_potential_layer, damage_potential_type, status_label):
        """This method synchronizes the damage potential with the EconoMe API."""

        # Check for valid user info
        if self.USER_INFO_VALIDATED:
            # Get damage potential layer
            DAMAGE_POTENTIAL_LAYER = damage_potential_layer
            # Check if damage potential layer is valid
            if DAMAGE_POTENTIAL_LAYER:
                # Damage potential type
                DAMAGE_POTENTIAL_TYPE = damage_potential_type
                DAMAGE_POTENTIAL_TYPE_DE = getattr(self.main_dialog, f"LABEL_DAM_POT_{DAMAGE_POTENTIAL_TYPE.upper()}").text()
                # Layer Asset list
                LAYER_ASSETS = {'AssetList': []}
                # Get current damage potential
                get_damage_potential = requests.get(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + '/damagepotential/asset', auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                UPLOADED_DAMAGE_POTENTIAL = json.loads(get_damage_potential.text)
                # Filter damage potential by object category
                OBJECT_CATEGORIES = [item for sublist in [list(self.ASSET_TYPEID_NAME_BY_CATEGORYNAME[obj_cat].values()) for obj_cat in self.ASSET_CATEGORIES_BY_DAMAGEPOTENTIAL_TYPE[DAMAGE_POTENTIAL_TYPE.upper()]] for item in sublist]
                FILTERED_UPLOADED_DAMAGE_POTENTIAL = [asset for asset in UPLOADED_DAMAGE_POTENTIAL if asset["AssetTypeID"] in OBJECT_CATEGORIES]
                # Retrieve parameter profile and get parameter associations
                get_parameter_profile = requests.get(self.ECONOME_API_PROFILE_BASE_PATH + f"/plugin-{DAMAGE_POTENTIAL_TYPE}", auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                PARAMETER_PROFILE = json.loads(get_parameter_profile.text)
                AssetIdentID_user_name = [pp["User"] for pp in PARAMETER_PROFILE['ParameterList'] if pp["API"].strip() == "AssetIdentID"][0]
                # Check for parameter profile
                if get_parameter_profile.status_code == 200:
                    # Get valid parameter pairs
                    damage_potential_valid_parameter_pairs = [pp for pp in PARAMETER_PROFILE['ParameterList'] if pp["User"].strip()]
                    damage_potential_valid_parameter_keys = [pp["API"] for pp in PARAMETER_PROFILE['ParameterList'] if pp["User"].strip()]
                    # Check if required fields were filled out
                    if set(self.DAMAGEPOTENTIAL_REQUIRED_FIELDS[damage_potential_type]).issubset(damage_potential_valid_parameter_keys):
                        layer_fields_names = [field.name() for field in DAMAGE_POTENTIAL_LAYER.fields()]
                        user_field_names = [pp["User"] for pp in damage_potential_valid_parameter_pairs]
                        # Check if the pp defined fields are present in the damage potential layer -> without this check, the user will not notice that the layer is not synchronizing correctly
                        if set(user_field_names).issubset(set(layer_fields_names)):
                            typechecked_user_field_names = [pp["User"] for pp in damage_potential_valid_parameter_pairs if DAMAGE_POTENTIAL_LAYER.fields().field(DAMAGE_POTENTIAL_LAYER.fields().indexFromName(pp["User"])).type() in self.QVARIANT_TYPES_DICT[pp["Type"]]]
                            # Check field types
                            if set(user_field_names).issubset(typechecked_user_field_names):
                                connected_columns_string = f"Verknüpfte Spalten: {len(typechecked_user_field_names)}: {', '.join(typechecked_user_field_names)}\n"
                                # Check that all features have an AssetIdentID
                                AssetIdentID_list = [feature[AssetIdentID_user_name] for feature in DAMAGE_POTENTIAL_LAYER.getFeatures()]
                                if all(AssetIdentID_list):
                                    # Create asset list from layer
                                    for feature in DAMAGE_POTENTIAL_LAYER.getFeatures():
                                        new_asset = {}
                                        for parameter_pair in damage_potential_valid_parameter_pairs:
                                            if parameter_pair["Type"] == "string":
                                                if feature[parameter_pair["User"]] == NULL:
                                                    new_asset[parameter_pair["API"]] = None
                                                elif isinstance(feature[parameter_pair["User"]], QVariant):
                                                    new_asset[parameter_pair["API"]] = str(feature[parameter_pair["User"]].value())
                                                else:
                                                    new_asset[parameter_pair["API"]] = str(feature[parameter_pair["User"]])
                                            elif parameter_pair["Type"] == "integer":
                                                if feature[parameter_pair["User"]] == NULL:
                                                    new_asset[parameter_pair["API"]] = None
                                                elif isinstance(feature[parameter_pair["User"]], QVariant):
                                                    new_asset[parameter_pair["API"]] = int(feature[parameter_pair["User"]].value())
                                                else:
                                                    new_asset[parameter_pair["API"]] = int(feature[parameter_pair["User"]])
                                            elif parameter_pair["Type"] == "float":
                                                if feature[parameter_pair["User"]] == NULL:
                                                    new_asset[parameter_pair["API"]] = None
                                                elif isinstance(feature[parameter_pair["User"]], QVariant):
                                                    new_asset[parameter_pair["API"]] = float(feature[parameter_pair["User"]].value())
                                                else:
                                                    new_asset[parameter_pair["API"]] = float(feature[parameter_pair["User"]])
                                            elif parameter_pair["Type"] == "boolean":
                                                if feature[parameter_pair["User"]] == NULL:
                                                    new_asset[parameter_pair["API"]] = None
                                                elif isinstance(feature[parameter_pair["User"]], QVariant):
                                                    new_asset[parameter_pair["API"]] = bool(feature[parameter_pair["User"]].value())
                                                else:
                                                    new_asset[parameter_pair["API"]] = bool(feature[parameter_pair["User"]])
                                        LAYER_ASSETS["AssetList"].append(new_asset)
                                    # Create list of asset ident ids
                                    layer_asset_ident_ids = [asset["AssetIdentID"] for asset in LAYER_ASSETS["AssetList"]]
                                    set_layer_asset_ident_ids = set(layer_asset_ident_ids)
                                    if len(layer_asset_ident_ids) == len(set_layer_asset_ident_ids):
                                        # Compare layer assets and uploaded damage potential
                                        comparison_result, mismatch_dict = self.compare_damage_potential(LAYER_ASSETS, FILTERED_UPLOADED_DAMAGE_POTENTIAL, damage_potential_valid_parameter_pairs)
                                        if comparison_result == "CANCEL":
                                            StatusDialog(status_icon = self.scaled_pixmap_warning, message_title=f"Fehler beim synchronisieren des Schadenpotenzials {DAMAGE_POTENTIAL_TYPE_DE}", message_text=f"\nKeine Option ausgewählt.\n").exec_()
                                            status_label.setPixmap(self.scaled_pixmap_warning)
                                            self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 0
                                        elif comparison_result == "POST LAYER ASSETS": # when no econome assets are present
                                            # Post new damage potential from layer
                                            post_damage_potential = requests.post(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + '/damagepotential/asset', json.dumps(LAYER_ASSETS), auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                                            if post_damage_potential.status_code == 200:
                                                created_assets = json.loads(post_damage_potential.text)["AssetList"]
                                                text = connected_columns_string
                                                text += f"\n{len(created_assets)} neue Schadenobjekte wurden erfolgreich vom Layer auf EconoMe hochgeladen:\n"
                                                for posted_asset_dict in created_assets:
                                                    text += f"{posted_asset_dict['AssetIdentID']}\n"
                                                StatusDialog(status_icon = self.scaled_pixmap_checkmark, message_title=f"Erfolg: Schadenpotenzial {DAMAGE_POTENTIAL_TYPE_DE} synchronisiert", message_text=text).exec_()
                                                status_label.setPixmap(self.scaled_pixmap_checkmark)
                                                self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 1
                                            else:
                                                StatusDialog(status_icon = self.scaled_pixmap_warning, message_title=f"Fehler beim synchronisieren des Schadenpotenzials {DAMAGE_POTENTIAL_TYPE_DE}", message_text=f"\nAPI-Fehlermeldung beim Hochladen des Schadenpotentials:\n{json.loads(post_damage_potential.text)}\n").exec_()
                                                status_label.setPixmap(self.scaled_pixmap_warning)
                                                self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 0
                                        # Layer assets match completely with assets in EconoMe
                                        elif comparison_result == "COMPLETE MATCH": # when all attributes match
                                            status_label.setPixmap(self.scaled_pixmap_checkmark)
                                            self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 1
                                            text = connected_columns_string
                                            text += f"\nLayer-Schadenobjekte stimmen mit EconoMe-Schadenobjekten überein.\n"
                                            StatusDialog(status_icon = self.scaled_pixmap_checkmark, message_title=f"Erfolg: Schadenpotenzial {DAMAGE_POTENTIAL_TYPE_DE} synchronisiert", message_text=text).exec_()
                                        # Update EconoMe assets
                                        elif comparison_result == "KEEP LAYER ASSETS": # when user decides to overwrite econome assets with layer assets
                                            """
                                            ATTENTION: 
                                            Only valid parameter pairs should be overwritten as the user might have changed attributes in EconoMe that are not present in the layer
                                            Therefore we put the layer attributes (instead of deleting and re-posting) into the EconoMe API.
                                            New layer assets are posted to EconoMe.
                                            """
                                            # 1) collect assets to patch/put/post in lists of dicts with correct API keys
                                            # 1.1) Assets where the type was changed in layer: we need to patch and then post them, thereby we need to preserve all EconoMe-sided stored attributes
                                            layer_assets_to_change_type = [layer_asset for layer_asset in LAYER_ASSETS["AssetList"] if layer_asset["AssetIdentID"] in mismatch_dict.keys() and "AssetTypeID" in mismatch_dict[layer_asset["AssetIdentID"]].keys() ]
                                            uploaded_assets_to_change_type = [asset for asset in FILTERED_UPLOADED_DAMAGE_POTENTIAL if asset["AssetIdentID"] in mismatch_dict.keys() and "AssetTypeID" in mismatch_dict[asset["AssetIdentID"]].keys()]
                                            patch_asset_ids = [asset["AssetID"] for asset in uploaded_assets_to_change_type]
                                            assets_to_change_type = []
                                            for uploaded_asset in copy.deepcopy(uploaded_assets_to_change_type):
                                                asset_to_change_type = uploaded_asset
                                                for key, value in mismatch_dict[uploaded_asset["AssetIdentID"]].items():
                                                    asset_to_change_type[key] = value["Layer"]
                                                asset_to_change_type.pop("AssetID") # not supported by API post
                                                asset_to_change_type.pop("ConnectedRouteID") # TODO report API bug
                                                assets_to_change_type.append(asset_to_change_type)
                                            # 1.2) Assets that exist in EconoMe: we need to update/put them
                                            layer_assets_to_put = [layer_asset for layer_asset in LAYER_ASSETS["AssetList"] if layer_asset["AssetIdentID"] in mismatch_dict.keys() and mismatch_dict[layer_asset["AssetIdentID"]] != {"Missing":"Fehlt in EconoMe."} and "AssetTypeID" not in mismatch_dict[layer_asset["AssetIdentID"]].keys()]
                                            assets_to_put = []
                                            for asset_to_put in copy.deepcopy(layer_assets_to_put):
                                                asset_to_put.pop("AssetTypeID") # not supported by API put (see above assets_to_change_type)
                                                asset_to_put.update({"AssetID": [asset["AssetID"] for asset in FILTERED_UPLOADED_DAMAGE_POTENTIAL if asset["AssetIdentID"] == asset_to_put["AssetIdentID"]][0]})
                                                asset_to_put.pop("AssetIdentID") # not supported by API put
                                                assets_to_put.append(asset_to_put)
                                            # 1.3) Assets that were newly created in the layer: we need to post them
                                            assets_to_post = [layer_asset for layer_asset in LAYER_ASSETS["AssetList"] if layer_asset["AssetIdentID"] in mismatch_dict.keys() and mismatch_dict[layer_asset["AssetIdentID"]] == {"Missing":"Fehlt in EconoMe."}]
                                            
                                            # 2) patch/put/post the assets
                                            # 2.1) Patch assets where the type was changed in layer
                                            if len(assets_to_change_type) != 0:
                                                patch_damage_potential = requests.patch(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + '/damagepotential/delete', json.dumps({"AssetIDList":patch_asset_ids}), auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                                                reput_damage_potential = requests.post(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + '/damagepotential/asset', json.dumps({"AssetList":assets_to_change_type}), auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                                                # Feedback to user about the syc errors or success
                                                if patch_damage_potential.status_code == 200 and reput_damage_potential.status_code == 200:
                                                    text = connected_columns_string
                                                    text += f"Der Objekttyp folgender {len(assets_to_change_type)} Schadenobjekte wurde auf EconoMe mit Objekttyp aus Layer überschrieben.\n"
                                                    text += f"Dafür musste jedes Schadenobjekt auf EconoMe gelöscht werden und neu erstellt werden. Alle Attribute wurden erhalten.\n"
                                                    for reput_asset_dict in json.loads(reput_damage_potential.text)["AssetList"]:
                                                        text += f"\n{reput_asset_dict['AssetIdentID']}"
                                                    StatusDialog(status_icon = self.scaled_pixmap_checkmark, message_title=f"Erfolg: Schadenpotenzial {DAMAGE_POTENTIAL_TYPE_DE} synchronisiert", message_text=text).exec_()
                                                    status_label.setPixmap(self.scaled_pixmap_checkmark)
                                                    self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 1
                                                else:
                                                    text = ""
                                                    if patch_damage_potential.status_code != 200:
                                                        text = f"\nAPI-Fehlermeldung beim Updaten des Objekttyps (Schritt 1 patch):\n{json.loads(patch_damage_potential.text)}\n"
                                                    elif reput_damage_potential.status_code != 200:
                                                        text += f"\nAPI-Fehlermeldung beim Updaten des Objekttyps (Schritt 2 post):\n{json.loads(reput_damage_potential.text)}\n"
                                                    StatusDialog(status_icon = self.scaled_pixmap_warning, message_title=f"Fehler beim synchronisieren des Schadenpotenzials {DAMAGE_POTENTIAL_TYPE_DE}", message_text=text).exec_()
                                                    status_label.setPixmap(self.scaled_pixmap_warning)
                                                    self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 0
                                            # 2.2) Put updated damage potential from layer
                                            if len(assets_to_put) != 0:
                                                put_damage_potential = requests.put(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + '/damagepotential/asset', json.dumps({"AssetList":assets_to_put}), auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                                                # Feedback to user about the syc errors or success
                                                if put_damage_potential.status_code == 200:
                                                    text = connected_columns_string
                                                    text += f"\nDie Attribute folgender {len(assets_to_put)} Schadenobjekte wurden auf EconoMe mit Attributen aus dem Layer überschrieben:\n"
                                                    #for put_asset_dict in json.loads(put_damage_potential.text)["AssetList"]: # TODO reportt bug an Peter: API does not return all updated assets
                                                    for put_asset_dict in layer_assets_to_put:
                                                        text += f"\n{put_asset_dict['AssetIdentID']}"
                                                    StatusDialog(status_icon = self.scaled_pixmap_checkmark, message_title=f"Erfolg: Schadenpotenzial {DAMAGE_POTENTIAL_TYPE_DE} synchronisiert", message_text=text).exec_()
                                                    status_label.setPixmap(self.scaled_pixmap_checkmark)
                                                    self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 1
                                                else:
                                                    StatusDialog(status_icon = self.scaled_pixmap_warning, message_title=f"Fehler beim synchronisieren des Schadenpotenzials {DAMAGE_POTENTIAL_TYPE_DE}", message_text=f"\nAPI-Fehlermeldung beim Updaten des vorhandenen Schadenpotentials (put):\n{json.loads(put_damage_potential.text)}\n").exec_()
                                                    status_label.setPixmap(self.scaled_pixmap_warning)
                                                    self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 0
                                            # 2.3) Post new damage potential from layer
                                            if len(assets_to_post) != 0:
                                                post_damage_potential = requests.post(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + '/damagepotential/asset', json.dumps({"AssetList":assets_to_post}), auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                                                # Feedback to user about the syc errors or success
                                                if post_damage_potential.status_code == 200:
                                                    # for newly created assets, we want to save the standard-created attributes from EconoMe to the layer
                                                    DAMAGE_POTENTIAL_LAYER.startEditing()
                                                    for posted_asset_dict in json.loads(post_damage_potential.text)["AssetList"]:
                                                        expression = f'"{AssetIdentID_user_name}" = \'{posted_asset_dict["AssetIdentID"]}\''
                                                        request = QgsFeatureRequest().setFilterExpression(expression)
                                                        feature = next(DAMAGE_POTENTIAL_LAYER.getFeatures(request), None)
                                                        for parameter_pair in damage_potential_valid_parameter_pairs:
                                                            feature.setAttribute(parameter_pair["User"], posted_asset_dict[parameter_pair["API"]])
                                                            DAMAGE_POTENTIAL_LAYER.updateFeature(feature)
                                                    DAMAGE_POTENTIAL_LAYER.commitChanges()
                                                    # user feedback
                                                    text = connected_columns_string
                                                    text += f"{len(assets_to_post)} neue Schadenobjekte wurden vom Layer auf EconoMe hochgeladen:\n"
                                                    for posted_asset_dict in json.loads(post_damage_potential.text)["AssetList"]:
                                                        text += f"\n{posted_asset_dict['AssetIdentID']}"
                                                    StatusDialog(status_icon = self.scaled_pixmap_checkmark, message_title=f"Erfolg: Schadenpotenzial {DAMAGE_POTENTIAL_TYPE_DE} synchronisiert", message_text=text).exec_()
                                                    status_label.setPixmap(self.scaled_pixmap_checkmark)
                                                    self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 1
                                                else:
                                                    StatusDialog(status_icon = self.scaled_pixmap_warning, message_title=f"Fehler beim synchronisieren des Schadenpotenzials {DAMAGE_POTENTIAL_TYPE_DE}", message_text=f"\nAPI-Fehlermeldung beim Hochladen des neuen Schadenpotentials (post):\n{json.loads(post_damage_potential.text)}\n").exec_()
                                                    status_label.setPixmap(self.scaled_pixmap_warning)
                                                    self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 0
                                            # Checkmark should only be changed if both requests were successful
                                            if len(assets_to_put) != 0 and len(assets_to_post) != 0:
                                                if post_damage_potential.status_code != 200 or put_damage_potential.status_code != 200:
                                                    status_label.setPixmap(self.scaled_pixmap_warning)
                                                    self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 0
                                        # Update Layer assets
                                        elif comparison_result == "KEEP ECONOME ASSETS":
                                            # Change layer assets based on EconoMe
                                            DAMAGE_POTENTIAL_LAYER.startEditing()
                                            for feature in DAMAGE_POTENTIAL_LAYER.getFeatures():
                                                AssetIdentID = feature[AssetIdentID_user_name]
                                                if AssetIdentID in mismatch_dict.keys():
                                                    updated_assetIdentIDs = []
                                                    deleted_assetIdentIDs = []
                                                    corresponding_uploaded_asset = [asset for asset in UPLOADED_DAMAGE_POTENTIAL if asset["AssetIdentID"] == str(feature[AssetIdentID_user_name])]
                                                    if len(corresponding_uploaded_asset) != 0:
                                                        for parameter_pair in damage_potential_valid_parameter_pairs:
                                                            feature.setAttribute(parameter_pair["User"], corresponding_uploaded_asset[0][parameter_pair["API"]])
                                                            DAMAGE_POTENTIAL_LAYER.updateFeature(feature)
                                                        updated_assetIdentIDs.append(AssetIdentID)
                                                    else:
                                                        DAMAGE_POTENTIAL_LAYER.dataProvider().deleteFeatures([feature.id()])
                                                        deleted_assetIdentIDs.append(AssetIdentID)
                                            DAMAGE_POTENTIAL_LAYER.commitChanges()
                                            text = connected_columns_string
                                            if len(updated_assetIdentIDs) != 0:
                                                text += f"\nDie Attribute folgender {len(updated_assetIdentIDs)} Schadenobjekte wurden im Layer mit Attributen aus EconoMe überschrieben:\n"
                                                for AssetIdentID in updated_assetIdentIDs:
                                                    text += f"\n{AssetIdentID}"
                                            if len(deleted_assetIdentIDs) != 0:
                                                text += f"\n\nFolgende {len(deleted_assetIdentIDs)} Schadenobjekte wurden im Layer gelöscht:\n"
                                                for AssetIdentID in deleted_assetIdentIDs:
                                                    text += f"\n{AssetIdentID}"
                                            StatusDialog(status_icon = self.scaled_pixmap_checkmark, message_title=f"Erfolg: Schadenpotenzial {DAMAGE_POTENTIAL_TYPE_DE} synchronisiert", message_text=text).exec_()
                                            status_label.setPixmap(self.scaled_pixmap_checkmark)
                                            self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 1
                                        elif comparison_result == "DELETE MISSING ASSETS":
                                            # Delete category-filtered damage potential assets from EconoMe
                                            patch_asset_ids = [asset["AssetID"] for asset in FILTERED_UPLOADED_DAMAGE_POTENTIAL if asset["AssetIdentID"] in mismatch_dict.keys() and mismatch_dict[asset["AssetIdentID"]] == {"Missing":"Fehlt im Layer."}]
                                            patch_assetIdentIDs = [asset["AssetIdentID"] for asset in FILTERED_UPLOADED_DAMAGE_POTENTIAL if asset["AssetIdentID"] in mismatch_dict.keys() and mismatch_dict[asset["AssetIdentID"]] == {"Missing":"Fehlt im Layer."}]
                                            patch_damage_potential = requests.patch(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + '/damagepotential/delete', json.dumps({"AssetIDList": patch_asset_ids}), auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                                            if patch_damage_potential.status_code == 200:
                                                text = connected_columns_string
                                                text += f"\nFolgende {len(patch_asset_ids)} Schadenobjekte wurden erfolgreich von EconoMe gelöscht:\n"
                                                for deleted_assetIdentID in patch_assetIdentIDs:
                                                    text += f"\n{deleted_assetIdentID}"
                                                text += f"\n\nEs wird noch nach Attribut-Konflikten gesucht..."
                                                StatusDialog(status_icon = self.scaled_pixmap_checkmark, message_title=f"Erfolg: Schadenpotenzial {DAMAGE_POTENTIAL_TYPE_DE} synchronisiert", message_text=text).exec_()
                                                self.synchronize_damage_potential(DAMAGE_POTENTIAL_LAYER, DAMAGE_POTENTIAL_TYPE, status_label)
                                        elif comparison_result == "NO CHANGE":
                                            # USER INFO: No changes made
                                            StatusDialog(status_icon = self.scaled_pixmap_warning, message_title="Hinweis", message_text=f"\nBitte fehlende Schadenobjekte im Layer hinzufügen!\n").exec_()
                                            status_label.setPixmap(self.scaled_pixmap_warning)
                                            self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 0
                                        else:
                                            # USER INFO: internal issue unknown option
                                            StatusDialog(status_icon = self.scaled_pixmap_warning, message_title="Interner Fehler", message_text=f"\nUnbekannte Option ausgewählt.\n").exec_()
                                    else:
                                        # USER INFO: Some of the layer features have the same AssetIdentID
                                        duplicate_AssetIdentIDs = [item for item, count in Counter(layer_asset_ident_ids).items() if count > 1]
                                        text = f"\nFolgende AssetIdentIDs kommen im Layer mehrmals vor:\n"
                                        for duplicate_AssetIdentID in duplicate_AssetIdentIDs:
                                            text += f"\n{duplicate_AssetIdentID}"
                                        text += f"\n\nBitte nur eindeutige AssetIdentIDs verwenden!\n"
                                        StatusDialog(status_icon = self.scaled_pixmap_warning, message_title=f"Fehler beim synchronisieren des Schadenpotenzials {DAMAGE_POTENTIAL_TYPE_DE}", message_text=text).exec_()
                                        status_label.setPixmap(self.scaled_pixmap_warning)
                                        self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 0
                                else:
                                    # USER INFO: Some of the layer features have no AssetIdentID / Null values for AssetIdentID
                                    assets_missing_AssetIdentID = [feature for feature in DAMAGE_POTENTIAL_LAYER.getFeatures() if not feature[AssetIdentID_user_name]]
                                    text = f"{len(assets_missing_AssetIdentID)} Schadenobjekte im Layer haben keine AssetIdentID! \nBitte AssetIdentID hinzufügen!\n"
                                    StatusDialog(status_icon = self.scaled_pixmap_warning, message_title=f"Fehler beim synchronisieren des Schadenpotenzials {DAMAGE_POTENTIAL_TYPE_DE}", message_text=text).exec_()
                                    status_label.setPixmap(self.scaled_pixmap_warning)
                                    self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 0
                            else:
                                # USER INFO: Some of the pp defined fields have a wrong data type
                                wrong_datatype_fields = set(user_field_names) - set(typechecked_user_field_names)
                                text = f"Folgende {len(wrong_datatype_fields)} Layer-Spalten haben einen falschen Datentyp:\n"
                                for field_name in wrong_datatype_fields:
                                    text += f"\n{field_name}"
                                text += f"\n\nBitte Parameter-Profil erneuern.\n"
                                StatusDialog(status_icon = self.scaled_pixmap_warning, message_title=f"Fehler beim synchronisieren des Schadenpotenzials {DAMAGE_POTENTIAL_TYPE_DE}", message_text=text).exec_()
                                status_label.setPixmap(self.scaled_pixmap_warning)
                                self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 0
                        else:
                            # USER INFO: Some of the pp defined fields are not present in the damage potential layer
                            missing_fields = set(user_field_names) - set(layer_fields_names)
                            text = f"Folgende {len(missing_fields)} Layer-Spalten wurden im Parameter-Profil einem Attribut zugewiesen, aber konnten im Layer nicht gefunden werden:\n"
                            for missing_field in missing_fields:
                                text += f"\n{missing_field}"
                            text += f"\n\nBitte Parameter-Profil erneuern.\n"
                            StatusDialog(status_icon = self.scaled_pixmap_warning, message_title=f"Fehler beim synchronisieren des Schadenpotenzials {DAMAGE_POTENTIAL_TYPE_DE}", message_text=text).exec_()
                            status_label.setPixmap(self.scaled_pixmap_warning)
                            self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 0
                    else:
                        # USER INFO: Not all required fields are defined in the parameter profile
                        missing_required_fields = set(self.DAMAGEPOTENTIAL_REQUIRED_FIELDS[damage_potential_type]) - set(damage_potential_valid_parameter_keys)
                        text = f"Folgenden {len(missing_required_fields)} Pflicht-Attributen wurden im Parameter-Profil keine Spalte des Layers zugewiesen:\n"
                        for required_field in missing_required_fields:
                            text += f"\n{required_field}"
                        text += f"\n\nBitte Parameter-Profil erneuern.\n"
                        StatusDialog(status_icon = self.scaled_pixmap_warning, message_title=f"Fehler beim synchronisieren des Schadenpotenzials {DAMAGE_POTENTIAL_TYPE_DE}", message_text=text).exec_()
                        status_label.setPixmap(self.scaled_pixmap_warning)
                        self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 0
                else:
                    # USER INFO: No parameter profile found
                    StatusDialog(status_icon = self.scaled_pixmap_warning, message_title=f"Fehler beim synchronisieren des Schadenpotenzials {DAMAGE_POTENTIAL_TYPE_DE}", message_text=f"\nBitte Parameter-Profil definieren: Jedem Attribut eine Attributspalte des Layer zuweisen!\n").exec_()
                    status_label.setPixmap(self.scaled_pixmap_warning)
                    self.DAMAGE_POTENTIAL_LAYER_STATUS[damage_potential_type.upper()] = 0
            else:
                # USER INFO: Invalid damage potential layer
                StatusDialog(status_icon = self.scaled_pixmap_warning, message_title="Fehler", message_text=f"\nBitte Schadenpotential-Layer auswählen!\n").exec_()
        else:
            # USER INFO: Invalid login info
            StatusDialog(status_icon = self.scaled_pixmap_warning, message_title="Fehler", message_text=f"\nUngültige Anmeldedaten!\n").exec_()

    def delete_damage_potential_assets(self, damage_potential_type, status_label):
        """This method deletes damage potential assets by object category."""
        
        # CHECK: EconoMe user info validated
        if self.USER_INFO_VALIDATED:
            # Get current damage potential
            get_damage_potential = requests.get(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + '/damagepotential/asset', auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
            UPLOADED_DAMAGE_POTENTIAL = json.loads(get_damage_potential.text)
            # CHECK: Existing damage potential in EconoMe
            if len(UPLOADED_DAMAGE_POTENTIAL) != 0:
                # Filter damage potential by object category
                OBJECT_CATEGORIES = [item for sublist in [list(self.ASSET_TYPEID_NAME_BY_CATEGORYNAME[obj_cat].values()) for obj_cat in self.ASSET_CATEGORIES_BY_DAMAGEPOTENTIAL_TYPE[damage_potential_type.upper()]] for item in sublist]
                FILTERED_UPLOADED_DAMAGE_POTENTIAL = [asset for asset in UPLOADED_DAMAGE_POTENTIAL if asset["AssetTypeID"] in OBJECT_CATEGORIES]
                # CHECK: Are there assets of the corresponding damage potential type ?
                if len(FILTERED_UPLOADED_DAMAGE_POTENTIAL) != 0:
                    DAMAGE_POTENTIAL_TYPE_DE = getattr(self.main_dialog, f"LABEL_DAM_POT_{damage_potential_type.upper()}").text()
                    are_you_sure = AreYouSure(f"\nSind Sie sicher, dass Sie alle Schadenobjekte der Objektkategorie '{DAMAGE_POTENTIAL_TYPE_DE}' in EconoMe löschen möchten ?\n")
                    user_choice = are_you_sure.exec_()
                    if user_choice == QDialog.Accepted:
                        # Delete filtered damage potential assets
                        patch_asset_ids = {"AssetIDList": [asset["AssetID"] for asset in FILTERED_UPLOADED_DAMAGE_POTENTIAL]}
                        patch_damage_potential = requests.patch(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + '/damagepotential/delete', json.dumps(patch_asset_ids), auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                        # SUCCESS
                        if patch_damage_potential.status_code == 200:
                            StatusDialog(status_icon = self.scaled_pixmap_checkmark, message_title = f"Erfolg: Schadenpotential {DAMAGE_POTENTIAL_TYPE_DE} erfolgreich gelöscht", message_text = f"\nAlle Schadenobjekte der Objektkategorie '{DAMAGE_POTENTIAL_TYPE_DE}' in EconoMe gelöscht!\n").exec_()
                            status_label.clear()
                        # ERROR
                        else:
                            StatusDialog(status_icon = self.scaled_pixmap_warning, message_title = f"Fehler beim Löschen des Schadenpotentials {DAMAGE_POTENTIAL_TYPE_DE}", message_text = f"\nAPI Fehlermeldung: {patch_damage_potential.json()}\n").exec_()
                else:
                    # USER INFO: Kein Schadenpotential vorhanden
                    StatusDialog(status_icon = self.scaled_pixmap_warning, message_title = f"Fehler beim Löschen des Schadenpotentials {damage_potential_type.upper()}", message_text = f"\nIn EconoMe ist kein Schadenpotential Objektkategorie '{damage_potential_type.upper()}' definiert!\n").exec_()
            else:
                # USER INFO: Kein Schadenpotential vorhanden
                StatusDialog(status_icon = self.scaled_pixmap_warning, message_title = f"Fehler beim Löschen des Schadenpotentials {damage_potential_type.upper()}", message_text = f"\nIn EconoMe ist kein Schadenpotential Objektkategorie '{damage_potential_type.upper()}' definiert!\n").exec_()

    def delete_damage_potential_affected_assets(self, annuality):
        """This method deletes affected damage potential assets by type."""
        
        # Filter scenes and get corresponding Scene ID
        get_scenes = requests.get(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + '/scene', auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
        SCENES = json.loads(get_scenes.text)
        process_ID = self.main_dialog.COMBO_BOX_HAZARD_PROCESS.currentText().split(", ")[0]
        variant_ID = self.main_dialog.COMBO_BOX_MEASURE_VARIANT.currentText().split(", ")[0]
        scene_ID = [scene["SceneID"] for scene in SCENES if int(scene["Annuality"]) == int(annuality) and int(scene["ProcessID"]) == int(process_ID) and int(scene["VariantID"]) == int(variant_ID)][0]
        # Get existing affected assets
        get_affected_assets = requests.get(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + f"/scene/{scene_ID}/asset", auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
        UPLOADED_AFFECTED_ASSETS = json.loads(get_affected_assets.text)
        # CHECK: Existing affected assets for specified scene ?
        if len(UPLOADED_AFFECTED_ASSETS) != 0:
            are_you_sure = AreYouSure(f"\nSind sie sicher, dass sie die Betroffenheit der Schadenpotentialobjekte für das {annuality}-jährliche Szenario löschen möchten ?\n")
            user_choice = are_you_sure.exec_()
            if user_choice == QDialog.Accepted:
                # Delete filtered damage potential assets
                delete_affected_assets = requests.delete(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + f"/scene/{scene_ID}/asset", auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                # SUCCESS
                if delete_affected_assets.status_code == 200:
                    StatusDialog(status_icon = self.scaled_pixmap_checkmark, message_title = "Erfolg", message_text = f"\nBetroffenheit der Schadenpotentialobjekte erfolgreich gelöscht!\n").exec_()
                # ERROR
                else:
                    StatusDialog(status_icon = self.scaled_pixmap_warning, message_title = "Fehler", message_text = f"\n{delete_affected_assets.json()}\n").exec_()
        else:
            # USER INFO: Keine AffectedAssets vorhanden
            StatusDialog(status_icon = self.scaled_pixmap_warning, message_title = "Fehler", message_text = f"\nKeine betroffenen Schadenpotentialobjekte definiert!\n").exec_()

    def get_hazard_intensity(self, damage_potential_layer, intensity_layer, intensity_layer_column, damage_potential_type):
        # For point features or buildings
        if damage_potential_layer.geometryType() == 0 or damage_potential_type == "BUILDING":
            # Define the parameters for the processing algorithm
            params = {
                'INPUT': damage_potential_layer,
                'JOIN': intensity_layer,
                'PREDICATE': [0],  # Spatial predicate: [0] means "intersects"
                'JOIN_FIELDS': [intensity_layer_column],
                'DISCARD_NONMATCHING': False,
                'PREFIX': '',
                'SUMMARIES': [3],  # Summary type: Maximum Value
                'OUTPUT': 'TEMPORARY_OUTPUT'
            }

            context = dataobjects.createContext()
            context.setInvalidGeometryCheck(QgsFeatureRequest.GeometryNoCheck)

            # Run the processing tool
            result = processing.run("qgis:joinbylocationsummary", params, context=context)

            # QgsProject.instance().addMapLayer(result['OUTPUT'])

            return result['OUTPUT']

        # For line features
        elif damage_potential_layer.geometryType() == 1:
            # Create a copy of the input damage potential layer
            type_and_projection = f"LineString?crs={damage_potential_layer.crs().postgisSrid()}"
            temp_layer = QgsVectorLayer(type_and_projection, "Temporary Layer", "memory")
            damage_potential_layer_fields = damage_potential_layer.fields()

            temp_layer.startEditing()

            temp_layer.dataProvider().addAttributes(damage_potential_layer_fields)

            # Copy all features from the original layer to the new layer
            for feature in damage_potential_layer.getFeatures():
                temp_layer.dataProvider().addFeature(feature)

            # Add new fields to the line layer to store overlay lengths
            temp_layer.dataProvider().addAttributes([
                QgsField('intensity_weak', QVariant.Double),
                QgsField('intensity_moderate', QVariant.Double),
                QgsField('intensity_strong', QVariant.Double)
            ])
            temp_layer.updateFields()

            # Create a dictionary to store lengths for each line feature
            line_lengths = {f.id(): {'low': 0, 'moderate': 0, 'high': 0} for f in temp_layer.getFeatures()}

            # Perform intersection and calculate lengths
            for line_feat in temp_layer.getFeatures():
                line_geom = line_feat.geometry()

                for poly_feat in intensity_layer.getFeatures():
                    poly_geom = poly_feat.geometry()
                    intensity = poly_feat[intensity_layer_column]

                    # Intersect line with polygon
                    intersection = line_geom.intersection(poly_geom)

                    if not intersection.isEmpty():
                        intensity_class = poly_feat[intensity_layer_column]
                        length = intersection.length()
                        if intensity_class == 1:
                            line_lengths[line_feat.id()]["low"] += length
                        elif intensity_class == 2:
                            line_lengths[line_feat.id()]["moderate"] += length
                        elif intensity_class == 3:
                            line_lengths[line_feat.id()]["high"] += length

            # Update the line layer with the calculated lengths
            for line_feat in temp_layer.getFeatures():
                line_feat['intensity_weak'] = line_lengths[line_feat.id()]['low']
                line_feat['intensity_moderate'] = line_lengths[line_feat.id()]['moderate']
                line_feat['intensity_strong'] = line_lengths[line_feat.id()]['high']
                temp_layer.updateFeature(line_feat)

            temp_layer.commitChanges()

            return temp_layer

        # For polygon features
        elif damage_potential_layer.geometryType() == 2:
            # Create a copy of the input damage potential layer
            type_and_projection = f"Polygon?crs={damage_potential_layer.crs().postgisSrid()}"
            temp_layer = QgsVectorLayer(type_and_projection, "Temporary Layer", "memory")
            damage_potential_layer_fields = damage_potential_layer.fields()

            temp_layer.startEditing()

            temp_layer.dataProvider().addAttributes(damage_potential_layer_fields)

            # Copy all features from the original layer to the new layer
            for feature in damage_potential_layer.getFeatures():
                temp_layer.dataProvider().addFeature(feature)

            # Add new fields to the line layer to store overlay lengths
            temp_layer.dataProvider().addAttributes([
                QgsField('intensity_weak', QVariant.Double),
                QgsField('intensity_moderate', QVariant.Double),
                QgsField('intensity_strong', QVariant.Double)
            ])
            temp_layer.updateFields()

            # Create a dictionary to store lengths for each line feature
            polygon_areas = {f.id(): {'low': 0, 'moderate': 0, 'high': 0} for f in temp_layer.getFeatures()}

            # Perform intersection and calculate lengths
            for poly_feat in temp_layer.getFeatures():
                poly_geom = poly_feat.geometry()
                for intensity_feature in intensity_layer.getFeatures():
                    intensity_geom = intensity_feature.geometry()
                    intensity = intensity_feature[intensity_layer_column]

                    # Intersect polygon with polygon
                    intersection = poly_geom.intersection(intensity_geom)

                    if not intersection.isEmpty():
                        intensity_class = intensity_feature[intensity_layer_column]
                        area = intersection.area()
                        if intensity_class == 1:
                            polygon_areas[poly_feat.id()]["low"] += area
                        elif intensity_class == 2:
                            polygon_areas[poly_feat.id()]["moderate"] += area
                        elif intensity_class == 3:
                            polygon_areas[poly_feat.id()]["high"] += area

            # Update the line layer with the calculated lengths
            for poly_feat in temp_layer.getFeatures():
                poly_feat['intensity_weak'] = polygon_areas[poly_feat.id()]['low']
                poly_feat['intensity_moderate'] = polygon_areas[poly_feat.id()]['moderate']
                poly_feat['intensity_strong'] = polygon_areas[poly_feat.id()]['high']
                temp_layer.updateFeature(poly_feat)

            temp_layer.commitChanges()

            return temp_layer

    def upload_consequence_analysis(self):
        """This method defines the affected assets and uploads the consequence analysis to EconoMe."""

        # CHECK: Valid user info
        if self.USER_INFO_VALIDATED:
            # DO: Create damage potential layer dictionary
            DAMAGE_POTENTIAL_LAYER_DICT = {
                "BUILDING": self.main_dialog.LAYER_DAM_POT_BUILDING.currentLayer(),
                "ROAD": self.main_dialog.LAYER_DAM_POT_ROAD.currentLayer(),
                "RAILWAY": self.main_dialog.LAYER_DAM_POT_RAILWAY.currentLayer(),
                "CABLE_TRANSPORT": self.main_dialog.LAYER_DAM_POT_CABLE_TRANSPORT.currentLayer(),
                "LINE_PIPE": self.main_dialog.LAYER_DAM_POT_LINE_PIPE.currentLayer(),
                "OTHER_ASSET": self.main_dialog.LAYER_DAM_POT_OTHER_ASSET.currentLayer()
            }
            # CHECK: One damage potential layer must be selected
            if any(layer is not None for layer in DAMAGE_POTENTIAL_LAYER_DICT.values()):
                # DO: Create a dictionary of all currently selected damage potential layers
                current_damage_potential_layers = {k: v for k, v in DAMAGE_POTENTIAL_LAYER_DICT.items() if v is not None}
                not_synced_damage_potential_layer_keys = [key for key in current_damage_potential_layers.keys() if self.DAMAGE_POTENTIAL_LAYER_STATUS[key] == 0]
                # CHECK: Selected damage potential layers are synced with EconoMe
                if len(not_synced_damage_potential_layer_keys) == 0:
                    # DO: Create a dictionary of all intensity layers, intensity columns and a placeholder for the intensity results
                    INTENSITY_LAYER_DICT = {
                        self.main_dialog.LABEL_SCENARIO_FREE.text().strip('-jährlich'): {
                            "intensity_layer": self.main_dialog.LAYER_INTENSITY_SCENARIO_FREE.currentLayer(),
                            "intensity_column": self.main_dialog.COMBO_BOX_INTENSITY_COLUMN_NAME_FREE.currentText(),
                            "intensity_results": None
                        },
                        "30": {
                            "intensity_layer": self.main_dialog.LAYER_INTENSITY_SCENARIO_30.currentLayer(),
                            "intensity_column": self.main_dialog.COMBO_BOX_INTENSITY_COLUMN_NAME_30.currentText(),
                            "intensity_results": None
                        },
                        "100": {
                            "intensity_layer": self.main_dialog.LAYER_INTENSITY_SCENARIO_100.currentLayer(),
                            "intensity_column": self.main_dialog.COMBO_BOX_INTENSITY_COLUMN_NAME_100.currentText(),
                            "intensity_results": None
                        },
                        "300": {
                            "intensity_layer": self.main_dialog.LAYER_INTENSITY_SCENARIO_300.currentLayer(),
                            "intensity_column": self.main_dialog.COMBO_BOX_INTENSITY_COLUMN_NAME_300.currentText(),
                            "intensity_results": None
                        },
                        self.main_dialog.LABEL_SCENARIO_EXTREME.text().strip('-jährlich'): {
                            "intensity_layer": self.main_dialog.LAYER_INTENSITY_SCENARIO_EXTREME.currentLayer(),
                            "intensity_column": self.main_dialog.COMBO_BOX_INTENSITY_COLUMN_NAME_EXTREME.currentText(),
                            "intensity_results": None
                        }
                    }
                    # DO: Get selection info
                    current_intensity_layer_info = {k: v for k, v in INTENSITY_LAYER_DICT.items() if v["intensity_layer"] is not None and v["intensity_column"] is not ""}
                    non_selected_intensity_layer_columns = {k: v for k, v in INTENSITY_LAYER_DICT.items() if v["intensity_layer"] is not None and v["intensity_column"] is ""}
                    # CHECK: One intensity layer and attribute column must be selected
                    if len(current_intensity_layer_info) != 0 and len(non_selected_intensity_layer_columns) == 0:
                        AFFECTED_ASSETS_SCENARIO_COLLECTION = {
                            self.main_dialog.LABEL_SCENARIO_FREE.text().strip('-jährlich'): [],
                            "30": [],
                            "100": [],
                            "300": [],
                            self.main_dialog.LABEL_SCENARIO_EXTREME.text().strip('-jährlich'): []
                        }
                        # Create a storage dictionary for errors
                        UPLOAD_CONSEQUENCE_ANALYSIS_ERRORS_DICT = {}
                        # Layer error
                        intensityLayerIsValid = True
                        # DO: Loop over selected damage potential layers
                        for damage_potential_type, current_damage_potential_layer in current_damage_potential_layers.items():
                            # DO: Get hazard intensities
                            if INTENSITY_LAYER_DICT[self.main_dialog.LABEL_SCENARIO_FREE.text().strip('-jährlich')]["intensity_layer"] and INTENSITY_LAYER_DICT[self.main_dialog.LABEL_SCENARIO_FREE.text().strip('-jährlich')]["intensity_column"]:
                                damage_potential_withIntensity_Free = self.get_hazard_intensity(
                                    current_damage_potential_layer,
                                    INTENSITY_LAYER_DICT[self.main_dialog.LABEL_SCENARIO_FREE.text().strip('-jährlich')]["intensity_layer"],
                                    INTENSITY_LAYER_DICT[self.main_dialog.LABEL_SCENARIO_FREE.text().strip('-jährlich')]["intensity_column"],
                                    damage_potential_type
                                )
                                INTENSITY_LAYER_DICT[self.main_dialog.LABEL_SCENARIO_FREE.text().strip('-jährlich')]["intensity_results"] = damage_potential_withIntensity_Free
                            if INTENSITY_LAYER_DICT["30"]["intensity_layer"] and INTENSITY_LAYER_DICT["30"]["intensity_column"]:
                                damage_potential_withIntensity_30 = self.get_hazard_intensity(
                                    current_damage_potential_layer,
                                    INTENSITY_LAYER_DICT["30"]["intensity_layer"],
                                    INTENSITY_LAYER_DICT["30"]["intensity_column"],
                                    damage_potential_type
                                )
                                INTENSITY_LAYER_DICT["30"]["intensity_results"] = damage_potential_withIntensity_30
                            if INTENSITY_LAYER_DICT["100"]["intensity_layer"] and INTENSITY_LAYER_DICT["100"]["intensity_column"]:
                                damage_potential_withIntensity_100 = self.get_hazard_intensity(
                                    current_damage_potential_layer,
                                    INTENSITY_LAYER_DICT["100"]["intensity_layer"],
                                    INTENSITY_LAYER_DICT["100"]["intensity_column"],
                                    damage_potential_type
                                )
                                INTENSITY_LAYER_DICT["100"]["intensity_results"] = damage_potential_withIntensity_100
                            if INTENSITY_LAYER_DICT["300"]["intensity_layer"] and INTENSITY_LAYER_DICT["300"]["intensity_column"]:
                                damage_potential_withIntensity_300 = self.get_hazard_intensity(
                                    current_damage_potential_layer,
                                    INTENSITY_LAYER_DICT["300"]["intensity_layer"],
                                    INTENSITY_LAYER_DICT["300"]["intensity_column"],
                                    damage_potential_type
                                )
                                INTENSITY_LAYER_DICT["300"]["intensity_results"] = damage_potential_withIntensity_300
                            if INTENSITY_LAYER_DICT[self.main_dialog.LABEL_SCENARIO_EXTREME.text().strip('-jährlich')]["intensity_layer"] and INTENSITY_LAYER_DICT[self.main_dialog.LABEL_SCENARIO_EXTREME.text().strip('-jährlich')]["intensity_column"]:
                                damage_potential_withIntensity_Extreme = self.get_hazard_intensity(
                                    current_damage_potential_layer,
                                    INTENSITY_LAYER_DICT[self.main_dialog.LABEL_SCENARIO_EXTREME.text().strip('-jährlich')]["intensity_layer"],
                                    INTENSITY_LAYER_DICT[self.main_dialog.LABEL_SCENARIO_EXTREME.text().strip('-jährlich')]["intensity_column"],
                                    damage_potential_type
                                )
                                INTENSITY_LAYER_DICT[self.main_dialog.LABEL_SCENARIO_EXTREME.text().strip('-jährlich')]["intensity_results"] = damage_potential_withIntensity_Extreme

                            # DO: Retrieve parameter profile and get parameter associations
                            get_parameter_profile = requests.get(self.ECONOME_API_PROFILE_BASE_PATH + f"/plugin-{damage_potential_type.lower()}", auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                            PARAMETER_PROFILE = json.loads(get_parameter_profile.text)

                            # DO: Get scenes
                            get_scenes = requests.get(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + '/scene', auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                            SCENES = json.loads(get_scenes.text)

                            # DO: Get damage potential
                            get_damage_potential = requests.get(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + '/damagepotential/asset', auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                            DAMAGE_POTENTIAL = json.loads(get_damage_potential.text)
                            # DO: Filter damage potential by object category
                            OBJECT_CATEGORIES = [item for sublist in [list(self.ASSET_TYPEID_NAME_BY_CATEGORYNAME[obj_cat].values()) for obj_cat in self.ASSET_CATEGORIES_BY_DAMAGEPOTENTIAL_TYPE[damage_potential_type]] for item in sublist]
                            FILTERED_DAMAGE_POTENTIAL = [asset for asset in DAMAGE_POTENTIAL if asset["AssetTypeID"] in OBJECT_CATEGORIES]

                            # DO: Create affected assets list for each scenario depending on damage potential layer type
                            for ann in [self.main_dialog.LABEL_SCENARIO_FREE.text().strip('-jährlich'), "30", "100", "300", self.main_dialog.LABEL_SCENARIO_EXTREME.text().strip('-jährlich')]:
                                # Create an empty affected assets list
                                AFFECTED_ASSETS = {
                                    "AffectedAssetList": []
                                }
                                # CHECK: Existing intensities
                                if INTENSITY_LAYER_DICT[ann]["intensity_results"]:
                                    
                                    # For BUILDING and OTHER ASSET point features
                                    if current_damage_potential_layer.geometryType() == 0 or damage_potential_type == "BUILDING":
                                        intensity_max_column = INTENSITY_LAYER_DICT[ann]["intensity_column"] + "_max"
                                        affected_features = [feature for feature in INTENSITY_LAYER_DICT[ann]["intensity_results"].getFeatures() if feature[intensity_max_column]]
                                        existing_special_parameter_fields_wrong_type = []
                                        for feature in affected_features:
                                            AssetIdentID_user_name = [p["User"] for p in PARAMETER_PROFILE['ParameterList'] if p["API"] == "AssetIdentID"][0]
                                            corresponding_asset = [asset for asset in FILTERED_DAMAGE_POTENTIAL if asset["AssetIdentID"] == str(feature[AssetIdentID_user_name])][0]

                                            intensity_class = None

                                            if int(feature[intensity_max_column]) == 1:
                                                intensity_class = "weak"
                                            elif int(feature[intensity_max_column]) == 2:
                                                intensity_class = "moderate"
                                            elif int(feature[intensity_max_column]) == 3:
                                                intensity_class = "strong"

                                            new_affected_asset = {
                                                "AssetID": corresponding_asset['AssetID'],
                                                "IntensityClass": intensity_class
                                            }

                                            ## Add special parameters
                                            # TEST: Layers with same projection
                                            affecting_intensity_area_test = [intens_feat for intens_feat in INTENSITY_LAYER_DICT[ann]["intensity_layer"].getFeatures() if intens_feat.geometry().intersects(feature.geometry())]
                                            if len(affecting_intensity_area_test) != 0:
                                                # Get corresponding intensity feature
                                                affecting_intensity_area = [intens_feat for intens_feat in INTENSITY_LAYER_DICT[ann]["intensity_layer"].getFeatures() if intens_feat.geometry().intersects(feature.geometry()) and int(intens_feat[INTENSITY_LAYER_DICT[ann]["intensity_column"]]) == int(feature[intensity_max_column])][0]
                                                # Get fields of intensity layer
                                                intens_layer_fields = INTENSITY_LAYER_DICT[ann]["intensity_layer"].fields()
                                                # Existing field names
                                                already_filled_parameters = []
                                                # Loop through special parameters list for object category
                                                for param in self.AFFECTED_ASSET_SPECIAL_PARAMETERS[damage_potential_type]:
                                                    # Check if parameter field exists in intensity layer
                                                    possible_fieldnames = [name for name in intens_layer_fields.names() if name in param[0]]
                                                    if len(possible_fieldnames) != 0:
                                                        if max(possible_fieldnames) not in already_filled_parameters:
                                                            # Get field name
                                                            fieldname = max(possible_fieldnames)
                                                            already_filled_parameters.append(fieldname)
                                                            # Get field
                                                            field = intens_layer_fields.field(fieldname)
                                                            # Write special parameters to the affected asset if type is correct
                                                            if param[1] == "string":
                                                                if field.type() == QVariant.String:
                                                                    if affecting_intensity_area[fieldname] is not None:
                                                                        new_affected_asset[param[0]] = affecting_intensity_area[fieldname]
                                                                else:
                                                                    existing_special_parameter_fields_wrong_type.append(param[0])
                                                            if param[1] == "integer":
                                                                if field.type() == QVariant.Int or field.type() == QVariant.LongLong:
                                                                    if affecting_intensity_area[fieldname] is not None:
                                                                        new_affected_asset[param[0]] = affecting_intensity_area[fieldname]
                                                                else:
                                                                    existing_special_parameter_fields_wrong_type.append(param[0])
                                                            if param[1] == "boolean":
                                                                if field.type() == QVariant.Bool:
                                                                    if affecting_intensity_area[fieldname] is not None:
                                                                        new_affected_asset[param[0]] = affecting_intensity_area[fieldname]
                                                                elif field.type() == QVariant.Int:
                                                                    if affecting_intensity_area[fieldname] is not None:
                                                                        if affecting_intensity_area[fieldname] == 0:
                                                                            new_affected_asset[param[0]] = False
                                                                        elif affecting_intensity_area[fieldname] == 1:
                                                                            new_affected_asset[param[0]] = True
                                                                else:
                                                                    existing_special_parameter_fields_wrong_type.append(param[0])
                                                            if param[1] == "number":
                                                                if field.type() == QVariant.Int or field.type() == QVariant.LongLong or field.type() == QVariant.Double:
                                                                    if affecting_intensity_area[fieldname] is not None:
                                                                        new_affected_asset[param[0]] = affecting_intensity_area[fieldname]
                                                                else:
                                                                    existing_special_parameter_fields_wrong_type.append(param[0])
                                                            if param[1] == "numZeroToOne":
                                                                if field.type() == QVariant.Int or field.type() == QVariant.LongLong or field.type() == QVariant.Double:
                                                                    if affecting_intensity_area[fieldname] is not None:
                                                                        if affecting_intensity_area[fieldname] >= 0 and affecting_intensity_area[fieldname] <= 1:
                                                                            new_affected_asset[param[0]] = affecting_intensity_area[fieldname]
                                                                else:
                                                                    existing_special_parameter_fields_wrong_type.append(param[0])
                                                        
                                                # USER INFO: Existing special parameter columns but wrong type
                                                if len(existing_special_parameter_fields_wrong_type) != 0:
                                                    StatusDialog(status_icon = self.scaled_pixmap_warning, message_title = "Fehler", message_text = f"\nFolgende Attributspalten spezieller Parameter haben den falschen Typ:\n-> {existing_special_parameter_fields_wrong_type}\n").exec_()
                                                    intensityLayerIsValid = False
                                                    break

                                                # Check for p_rA and missing p_rA_Comment
                                                if "p_rA" in new_affected_asset:
                                                    if "p_rA_Comment" in new_affected_asset:
                                                        if not new_affected_asset["p_rA_Comment"]:
                                                            new_affected_asset["p_rA_Comment"] = "-"       
                                                    else:
                                                        new_affected_asset["p_rA_Comment"] = "-" 
                                                        
                                                AFFECTED_ASSETS['AffectedAssetList'].append(new_affected_asset)
                                            # Display warning message, layers with different projections
                                            else:
                                                StatusDialog(status_icon = self.scaled_pixmap_warning, message_title="Fehler", message_text=f"\nSchadenpotential-Layer und Intensitätskarten mit unterschiedlichem Koordinatensystem. Bitte anpassen!\n").exec_()
                                                intensityLayerIsValid = False
                                                break
                                    # For ROAD, RAILWAY, CABLE TRANSPORT and OTHER ASSET line or polygon features
                                    elif current_damage_potential_layer.geometryType() == 1 or current_damage_potential_layer.geometryType() == 2:
                                        affected_features = [feature for feature in INTENSITY_LAYER_DICT[ann]["intensity_results"].getFeatures() if feature['intensity_weak'] > 0 or feature['intensity_moderate'] > 0 or feature['intensity_strong'] > 0]
                                        existing_special_parameter_fields_wrong_type = []
                                        for feature in affected_features:
                                            AssetIdentID_user_name = [p["User"] for p in PARAMETER_PROFILE['ParameterList'] if p["API"] == "AssetIdentID"][0]
                                            corresponding_asset = [asset for asset in FILTERED_DAMAGE_POTENTIAL if asset["AssetIdentID"] == str(feature[AssetIdentID_user_name])][0]

                                            intensity_weak = 0
                                            intensity_moderate = 0
                                            intensity_strong = 0

                                            if feature['intensity_weak']:
                                                intensity_weak = feature['intensity_weak']
                                            if feature['intensity_moderate']:
                                                intensity_moderate = feature['intensity_moderate']
                                            if feature['intensity_strong']:
                                                intensity_strong = feature['intensity_strong']

                                            new_affected_asset = {
                                                "AssetID": corresponding_asset['AssetID'],
                                                "WeakIntensity": intensity_weak,
                                                "ModerateIntensity": intensity_moderate,
                                                "StrongIntensity": intensity_strong
                                            }

                                            ## Add special parameters (Methodik wird geprüft!)

                                            """
                                            # Get corresponding intensity feature
                                            affecting_intensity_pieces = [intens_feat for intens_feat in INTENSITY_LAYER_DICT[ann]["intensity_layer"].getFeatures() if intens_feat.geometry().intersects(feature.geometry())]
                                            # Get fields of intensity layer
                                            intens_layer_fields = INTENSITY_LAYER_DICT[ann]["intensity_layer"].fields()
                                            # Existing field names
                                            already_filled_parameters = []
                                            # Loop through special parameters list for object category
                                            for param in self.AFFECTED_ASSET_SPECIAL_PARAMETERS[damage_potential_type]:
                                                # Check if parameter field exists in intensity layer
                                                possible_fieldnames = [name for name in intens_layer_fields.names() if name in param[0]]
                                                if len(possible_fieldnames) != 0:
                                                    if max(possible_fieldnames) not in already_filled_parameters:
                                                        # Get field name
                                                        fieldname = max(possible_fieldnames)
                                                        already_filled_parameters.append(fieldname)
                                                        # Get field
                                                        field = intens_layer_fields.field(fieldname)
                                                        affecting_intensity_collect_values = [feat[fieldname] for feat in affecting_intensity_pieces]
                                                        # Write special parameters to the affected asset if type is correct
                                                        if param[1] == "string":
                                                            if field.type() == QVariant.String:
                                                                concated_comments = ""
                                                                for val in affecting_intensity_collect_values:
                                                                    if val and val not in concated_comments:
                                                                        concated_comments += val
                                                                new_affected_asset[param[0]] = concated_comments
                                                            else:
                                                                existing_special_parameter_fields_wrong_type.append(param[0])
                                                        if param[1] == "integer":
                                                            if field.type() == QVariant.Int or field.type() == QVariant.LongLong:
                                                                if len(affecting_intensity_collect_values) != 0:
                                                                    for val in affecting_intensity_collect_values:
                                                                        if val:
                                                                            new_affected_asset[param[0]] = val
                                                                            break
                                                            else:
                                                                existing_special_parameter_fields_wrong_type.append(param[0])
                                                        if param[1] == "boolean":
                                                            if field.type() == QVariant.Bool:
                                                                if len(affecting_intensity_collect_values) != 0:
                                                                    for val in affecting_intensity_collect_values:
                                                                        if val:
                                                                            new_affected_asset[param[0]] = val
                                                                            break
                                                            elif field.type() == QVariant.Int:
                                                                if len(affecting_intensity_collect_values) != 0:
                                                                    for val in affecting_intensity_collect_values:
                                                                        if val == 1:
                                                                            new_affected_asset[param[0]] = val
                                                                            break
                                                            else:
                                                                existing_special_parameter_fields_wrong_type.append(param[0])
                                                        if param[1] == "number":
                                                            if field.type() == QVariant.Int or field.type() == QVariant.LongLong or field.type() == QVariant.Double:
                                                                max_val = 0
                                                                for val in affecting_intensity_collect_values:
                                                                    if val and val > max_val:
                                                                        max_val = val
                                                                new_affected_asset[param[0]] = max_val
                                                            else:
                                                                existing_special_parameter_fields_wrong_type.append(param[0])
                                                        if param[1] == "numZeroToOne":
                                                            if field.type() == QVariant.Int or field.type() == QVariant.LongLong or field.type() == QVariant.Double:
                                                                max_val = 0
                                                                for val in affecting_intensity_collect_values:
                                                                    if val and val > max_val:
                                                                        max_val = val
                                                                if max_val >= 0 and max_val <= 1:
                                                                    new_affected_asset[param[0]] = max_val
                                                            else:
                                                                existing_special_parameter_fields_wrong_type.append(param[0])
                                                    
                                            # USER INFO: Existing special parameter columns but wrong type
                                            if len(existing_special_parameter_fields_wrong_type) != 0:
                                                StatusDialog(status_icon = self.scaled_pixmap_warning, message_title = "Fehler", message_text = f"\nFolgende Attributspalten spezieller Parameter haben den falschen Typ:\n{existing_special_parameter_fields_wrong_type}\n").exec_()
                                                intensityLayerIsValid = False
                                                break

                                            # Check for p_rA and missing p_rA_Comment
                                            if "p_rA" in new_affected_asset:
                                                if "p_rA_Comment" in new_affected_asset:
                                                    if not new_affected_asset["p_rA_Comment"]:
                                                        new_affected_asset["p_rA_Comment"] = "n/a"       
                                                else:
                                                    StatusDialog(status_icon = self.scaled_pixmap_warning, message_title = "Fehler", message_text = f"\nBei Mitgabe der p_rA-Spalte, braucht es auch eine p_rA_Comment!\n").exec_()
                                                    intensityLayerIsValid = False
                                                    break

                                            # Check for p_vSp and missing p_vSp_Comment
                                            if "p_vSp" in new_affected_asset:
                                                if "p_vSp_Comment" in new_affected_asset:
                                                    if not new_affected_asset["p_vSp_Comment"]:
                                                        new_affected_asset["p_vSp_Comment"] = "n/a"       
                                                else:
                                                    StatusDialog(status_icon = self.scaled_pixmap_warning, message_title = "Fehler", message_text = f"\nBei Mitgabe der p_vSp-Spalte, braucht es auch eine p_vSp_Comment-Spalte!\n").exec_()
                                                    intensityLayerIsValid = False
                                                    break
                                                    
                                            """
                                                    
                                            # StatusDialog(f"\n{new_affected_asset}\n").exec_()
                                            AFFECTED_ASSETS['AffectedAssetList'].append(new_affected_asset)

                                if len(AFFECTED_ASSETS['AffectedAssetList']) != 0:
                                    AFFECTED_ASSETS_SCENARIO_COLLECTION[ann].extend(AFFECTED_ASSETS['AffectedAssetList'])

                        for ann in [self.main_dialog.LABEL_SCENARIO_FREE.text().strip('-jährlich'), "30", "100", "300", self.main_dialog.LABEL_SCENARIO_EXTREME.text().strip('-jährlich')]:
                            # CHECK: Is the AffectedAssetList empty ?
                            if AFFECTED_ASSETS_SCENARIO_COLLECTION[ann] and intensityLayerIsValid:
                                # StatusDialog("", f"{AFFECTED_ASSETS_SCENARIO_COLLECTION[ann]}").exec_()
                                AFFECTED_ASSETS_PER_SCENARIO = {
                                    'AffectedAssetList': AFFECTED_ASSETS_SCENARIO_COLLECTION[ann]
                                }
                                # DO: Get corresponding scene ID based on annuality, selected process IDs and variant IDs
                                process_ID = self.main_dialog.COMBO_BOX_HAZARD_PROCESS.currentText().split(", ")[0]
                                variant_ID = 0
                                if self.main_dialog.COMBO_BOX_MEASURE_VARIANT.currentText().split(", ")[0]:
                                    variant_ID = self.main_dialog.COMBO_BOX_MEASURE_VARIANT.currentText().split(",")[0]
                                scenes_match = [scene["SceneID"] for scene in SCENES if int(scene["Annuality"]) == int(ann) and int(scene["ProcessID"]) == int(process_ID) and int(scene["VariantID"]) == int(variant_ID)]
                                if len(scenes_match) != 0:
                                    scene_ID = scenes_match[0]
                                    # DO: Get existing affected assets
                                    get_affected_assets = requests.get(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + f"/scene/{scene_ID}/asset", auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                                    UPLOADED_AFFECTED_ASSETS = json.loads(get_affected_assets.text)
                                    # Existing affected assets, update affected assets
                                    if len(UPLOADED_AFFECTED_ASSETS) != 0:

                                        delete_affected_assets = requests.delete(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + f"/scene/{scene_ID}/asset", auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                                        post_affected_assets = requests.post(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + f"/scene/{scene_ID}/asset", json.dumps(AFFECTED_ASSETS_PER_SCENARIO), auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))

                                        if post_affected_assets.status_code != 200:
                                            UPLOAD_CONSEQUENCE_ANALYSIS_ERRORS_DICT[ann] = json.loads(post_affected_assets.text)
                                    # No existing affected assets, post new affected assets
                                    else:
                                        post_affected_assets = requests.post(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + f"/scene/{scene_ID}/asset", json.dumps(AFFECTED_ASSETS_PER_SCENARIO), auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                                        if post_affected_assets.status_code != 200:
                                            UPLOAD_CONSEQUENCE_ANALYSIS_ERRORS_DICT[ann] = json.loads(post_affected_assets.text)

                        # USER INFO
                        if len(UPLOAD_CONSEQUENCE_ANALYSIS_ERRORS_DICT) == 0 and intensityLayerIsValid:
                            StatusDialog(status_icon = self.scaled_pixmap_checkmark, message_title="Erfolg", message_text=f"\nKonsequenzanalyse erfolgreich hochgeladen!\n").exec_()
                        else:
                            StatusDialog(status_icon = self.scaled_pixmap_warning, message_title="Fehler", message_text=f"\nFehler beim Hochladen der Konsequenzanalyse!\n").exec_()
                    else:
                        # USER INFO: No intensity layer + attribute column selected
                        StatusDialog(status_icon = self.scaled_pixmap_warning, message_title="Fehler", message_text=f"\nBitte Intensitäts-Layer mit (!) Intensitätsfeld ausgewählen!\n").exec_()
                else:
                    # USER INFO: At least one selected damage potential layer is not synced with EconoMe
                    StatusDialog(status_icon = self.scaled_pixmap_warning, message_title="Fehler", message_text=f"\nNicht alle Schadenpotential-Layers sind bereit für den Verschnitt!\nBitte synchronisiere die ausgewählten Schadenpotential-Layer oder wähle sie ab, um fortzufahren!\n").exec_()
            else:
                # USER INFO: No damage potential layer selected
                StatusDialog(status_icon = self.scaled_pixmap_warning, message_title="Fehler", message_text=f"\nKein Schadenpotential-Layer ausgewählt!\n").exec_()
        else:
            # USER INFO: No valid user info
            StatusDialog(status_icon = self.scaled_pixmap_warning, message_title="Fehler", message_text=f"\nUngültige Anmeldedaten: Projektnummer oder API-Keys!\n").exec_()

    def visualize_risks(self):
        """Fetches risk from EconoMe and creates a visualized layer."""

        # CHECK: Valid user info
        if self.USER_INFO_VALIDATED:
            # Create damage potential layer list
            DAMAGE_POTENTIAL_LAYER_DICT = {
                "BUILDING": self.main_dialog.LAYER_DAM_POT_BUILDING.currentLayer(),
                "ROAD": self.main_dialog.LAYER_DAM_POT_ROAD.currentLayer(),
                "RAILWAY": self.main_dialog.LAYER_DAM_POT_RAILWAY.currentLayer(),
                "CABLE_TRANSPORT": self.main_dialog.LAYER_DAM_POT_CABLE_TRANSPORT.currentLayer(),
                "LINE_PIPE": self.main_dialog.LAYER_DAM_POT_LINE_PIPE.currentLayer(),
                "OTHER_ASSET": self.main_dialog.LAYER_DAM_POT_OTHER_ASSET.currentLayer()
            }
            # CHECK: One damage potential layer selected
            if any(layer is not None for layer in DAMAGE_POTENTIAL_LAYER_DICT.values()):
                # Create a dictionary of all input damage potential layers
                current_damage_potential_layers = {k: v for k, v in DAMAGE_POTENTIAL_LAYER_DICT.items() if v is not None}
                not_synced_damage_potential_layer_keys = [key for key in current_damage_potential_layers.keys() if self.DAMAGE_POTENTIAL_LAYER_STATUS[key] == 0]
                # CHECK: Selected damage potential layers are synced with EconoMe
                if len(not_synced_damage_potential_layer_keys) == 0:
                    # Loop over damage potential layers
                    for key, current_damage_potential_layer in current_damage_potential_layers.items():
                        # Get corresponding parameter profile
                        get_parameter_profile = requests.get(self.ECONOME_API_PROFILE_BASE_PATH + f"/plugin-{key.lower()}", auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                        PARAMETER_PROFILE = json.loads(get_parameter_profile.text)
                        # Get hazard processes
                        get_hazard_processes = requests.get(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + "/process", auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                        HAZARD_PROCESSES = json.loads(get_hazard_processes.text)
                        SELECTED_HAZARD_PROCESS = self.main_dialog.COMBO_BOX_HAZARD_PROCESS.currentText()
                        # Get measure variant
                        SELECTED_MEASURE_VARIANT = self.main_dialog.COMBO_BOX_MEASURE_VARIANT.currentText()
                        # Get damage potential
                        get_damage_potential = requests.get(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + "/damagepotential/asset", auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                        UPLOADED_DAMAGE_POTENTIAL = json.loads(get_damage_potential.text)
                        # Get scenes
                        get_scenes = requests.get(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + "/scene", auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                        SCENES = json.loads(get_scenes.text)
                        # CHECK: Existing scenes
                        if len(SCENES) != 0:

                            free_annuality = self.main_dialog.LABEL_SCENARIO_FREE.text().strip('-jährlich').strip()
                            extreme_annuality = self.main_dialog.LABEL_SCENARIO_EXTREME.text().strip('-jährlich').strip()

                            # Store risk data output
                            RISK_OUTPUT_DICT = {
                                free_annuality: {
                                    "risk": None
                                },
                                "30": {
                                    "risk": None
                                },
                                "100": {
                                    "risk": None
                                },
                                "300": {
                                    "risk": None
                                },
                                extreme_annuality: {
                                    "risk": None
                                }
                            }

                            # Get risk from all scenes matching user selection

                            # AnnualityFree
                            filter_scenes = [scene["SceneID"] for scene in SCENES if scene["Annuality"] == int(free_annuality) and int(scene["ProcessID"]) == int(SELECTED_HAZARD_PROCESS.split(", ")[0]) and int(scene["VariantID"]) == int(SELECTED_MEASURE_VARIANT.split(", ")[0])]
                            if filter_scenes:
                                scene_ID = filter_scenes[0]
                                get_risk_free = requests.get(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + f"/risk/scene/{scene_ID}", auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                                RISK_OUTPUT_DICT[free_annuality]["risk"] = json.loads(get_risk_free.text)
                            # Annuality30
                            filter_scenes = [scene["SceneID"] for scene in SCENES if scene["Annuality"] == 30 and int(scene["ProcessID"]) == int(SELECTED_HAZARD_PROCESS.split(", ")[0]) and int(scene["VariantID"]) == int(SELECTED_MEASURE_VARIANT.split(", ")[0])]
                            if filter_scenes:
                                scene_ID = filter_scenes[0]
                                get_risk_30 = requests.get(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + f"/risk/scene/{scene_ID}", auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                                RISK_OUTPUT_DICT["30"]["risk"] = json.loads(get_risk_30.text)
                            # Annuality100
                            filter_scenes = [scene["SceneID"] for scene in SCENES if scene["Annuality"] == 100 and int(scene["ProcessID"]) == int(SELECTED_HAZARD_PROCESS.split(", ")[0]) and int(scene["VariantID"]) == int(SELECTED_MEASURE_VARIANT.split(", ")[0])]
                            if filter_scenes:
                                scene_ID = filter_scenes[0]
                                get_risk_100 = requests.get(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + f"/risk/scene/{scene_ID}", auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                                RISK_OUTPUT_DICT["100"]["risk"] = json.loads(get_risk_100.text)
                            # Annuality100
                            filter_scenes = [scene["SceneID"] for scene in SCENES if scene["Annuality"] == 300 and int(scene["ProcessID"]) == int(SELECTED_HAZARD_PROCESS.split(", ")[0]) and int(scene["VariantID"]) == int(SELECTED_MEASURE_VARIANT.split(", ")[0])]
                            if filter_scenes:
                                scene_ID = filter_scenes[0]
                                get_risk_300 = requests.get(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + f"/risk/scene/{scene_ID}", auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                                RISK_OUTPUT_DICT["300"]["risk"] = json.loads(get_risk_300.text)
                            # AnnualityExtreme
                            filter_scenes = [scene["SceneID"] for scene in SCENES if scene["Annuality"] == int(extreme_annuality) and int(scene["ProcessID"]) == int(SELECTED_HAZARD_PROCESS.split(", ")[0]) and int(scene["VariantID"]) == int(SELECTED_MEASURE_VARIANT.split(", ")[0])]
                            if filter_scenes:
                                scene_ID = filter_scenes[0]
                                get_risk_extreme = requests.get(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + f"/risk/scene/{scene_ID}", auth=HTTPBasicAuth(self.ECONOME_USER_NAME, self.ECONOME_PASSWORD))
                                RISK_OUTPUT_DICT[extreme_annuality]["risk"] = json.loads(get_risk_extreme.text)

                            has_risk = any(val["risk"] not in (None, [])  for val in RISK_OUTPUT_DICT.values())
                            # CHECK: Are there risks in EconoMe
                            if has_risk:
                                # Create a copy of the damage potential layer as a memory layer with the prefix of the hazard process
                                # "Point?crs=EPSG:2056"
                                type_and_projection = ""
                                if current_damage_potential_layer.geometryType() == 0:
                                    # layer.geometryType() 0,1,2 Point Line Polygon  ---- layer.crs().postgisSrid()
                                    type_and_projection = f"Point?crs={current_damage_potential_layer.crs().postgisSrid()}"
                                elif current_damage_potential_layer.geometryType() == 1:
                                    # layer.geometryType() 0,1,2 Point Line Polygon  ---- layer.crs().postgisSrid()
                                    type_and_projection = f"LineString?crs={current_damage_potential_layer.crs().postgisSrid()}"
                                elif current_damage_potential_layer.geometryType() == 2:
                                    # layer.geometryType() 0,1,2 Point Line Polygon  ---- layer.crs().postgisSrid()
                                    type_and_projection = f"Polygon?crs={current_damage_potential_layer.crs().postgisSrid()}"

                                current_damage_potential_layer_AssetIdentID = [p["User"] for p in PARAMETER_PROFILE['ParameterList'] if p["API"] == "AssetIdentID"][0]
                                current_damage_potential_layer_fields = current_damage_potential_layer.fields()
                                timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
                                new_risk_layer = QgsVectorLayer(type_and_projection, f"Risiken_{SELECTED_MEASURE_VARIANT.split(', ')[1][:25]}_{SELECTED_HAZARD_PROCESS.split(', ')[2][:25]}_{key}_{timestamp}", "memory")
                                new_risk_layer.dataProvider().addAttributes(current_damage_potential_layer_fields)

                                # Switch to edit mode
                                new_risk_layer.startEditing()

                                # Copy all features from the original layer to the new layer
                                for feature in current_damage_potential_layer.getFeatures():
                                    new_risk_layer.dataProvider().addFeature(feature)

                                # Loop over all risks and define new fields
                                for annuality in RISK_OUTPUT_DICT.keys():
                                    # CHECK: Is there risk data ?
                                    if RISK_OUTPUT_DICT[annuality]["risk"]:
                                        # Define new fields if not existing
                                        for output_variable in self.RISK_OUTPUT_VARIABLES[key]:
                                            if output_variable not in current_damage_potential_layer_fields:
                                                new_field = QgsField(f"{output_variable}_{annuality}", QVariant.Double)
                                                new_risk_layer.dataProvider().addAttributes([new_field])
                                        # Update fields
                                        new_risk_layer.updateFields()

                                # Loop over all risks and fill fields
                                for annuality in RISK_OUTPUT_DICT.keys():
                                    # CHECK: Is there risk data ?
                                    if RISK_OUTPUT_DICT[annuality]["risk"]:
                                        # LOOP over all features
                                        for feature in new_risk_layer.getFeatures():
                                            # Get corresponding Asset ID
                                            corresponding_asset_id = [asset["AssetID"] for asset in UPLOADED_DAMAGE_POTENTIAL if asset["AssetIdentID"] == str(feature[current_damage_potential_layer_AssetIdentID])][0]
                                            result = [elem for elem in RISK_OUTPUT_DICT[annuality]["risk"] if int(elem["AssetID"]) == int(corresponding_asset_id)]
                                            if len(result) != 0:
                                                # Loop over all output variables
                                                for output_variable in self.RISK_OUTPUT_VARIABLES[key]:
                                                    # Enter attributes
                                                    feature.setAttribute(f"{output_variable}_{annuality}", result[0][output_variable])
                                            # Update feature
                                            new_risk_layer.updateFeature(feature)

                                # Save edits
                                new_risk_layer.commitChanges()
                                # Set color renderer

                                new_risk_layer_fields = [field.name() for field in new_risk_layer.fields()]

                                # Add summary risk fields
                                if f"R_{free_annuality}" in new_risk_layer_fields or "R_30" in new_risk_layer_fields or "R_100" in new_risk_layer_fields or "R_300" in new_risk_layer_fields or f"R_{extreme_annuality}" in new_risk_layer_fields:
                                    if key not in ["OTHER_ASSET", "LINE_PIPE"]:
                                        new_field = QgsField("R", QVariant.Double)
                                        new_risk_layer.dataProvider().addAttributes([new_field])
                                        new_field = QgsField("r_indiv", QVariant.Double)
                                        new_risk_layer.dataProvider().addAttributes([new_field])
                                        # Update fields
                                        new_risk_layer.updateFields()

                                new_risk_layer_fields = [field.name() for field in new_risk_layer.fields()]
                                new_risk_layer.startEditing()

                                # Add summary risk fields
                                for feature in new_risk_layer.getFeatures():
                                    r_sum = 0
                                    if "R" in new_risk_layer_fields:
                                        r_sum = 0
                                        if f"R_{free_annuality}" in new_risk_layer_fields:
                                            if feature[f"R_{free_annuality}"] is not None:
                                                r_sum += feature[f"R_{free_annuality}"]
                                        if "R_30" in new_risk_layer_fields:
                                            if feature["R_30"] is not None:
                                                r_sum += feature["R_30"]
                                        if "R_100" in new_risk_layer_fields:
                                            if feature["R_100"] is not None:
                                                r_sum += feature["R_100"]
                                        if "R_300" in new_risk_layer_fields:
                                            if feature["R_300"] is not None:
                                                r_sum += feature["R_300"]
                                        if f"R_{free_annuality}" in new_risk_layer_fields:
                                            if feature[f"R_{extreme_annuality}"] is not None:
                                                r_sum += feature[f"R_{extreme_annuality}"]

                                        feature.setAttribute("R", r_sum)

                                    r_indiv_sum = 0
                                    if "r_indiv" in new_risk_layer_fields:
                                        r_indiv_sum = 0
                                        if f"r_indiv_{free_annuality}" in new_risk_layer_fields:
                                            if feature[f"r_indiv_{free_annuality}"] is not None:
                                                r_indiv_sum += feature[f"r_indiv_{free_annuality}"]
                                        if f"r_indiv_30" in new_risk_layer_fields:
                                            if feature["r_indiv_30"] is not None:
                                                r_indiv_sum += feature["r_indiv_30"]
                                        if f"r_indiv_100" in new_risk_layer_fields:
                                            if feature["r_indiv_100"] is not None:
                                                r_indiv_sum += feature["r_indiv_100"]
                                        if f"r_indiv_300" in new_risk_layer_fields:
                                            if feature["r_indiv_300"] is not None:
                                                r_indiv_sum += feature["r_indiv_300"]
                                        if f"r_indiv_{extreme_annuality}" in new_risk_layer_fields:
                                            if feature[f"r_indiv_{extreme_annuality}"] is not None:
                                                r_indiv_sum += feature[f"r_indiv_{extreme_annuality}"]

                                        feature.setAttribute("r_indiv", r_indiv_sum)

                                    new_risk_layer.updateFeature(feature)

                                new_risk_layer.commitChanges()

                                # Visualize r_indiv
                                if "r_indiv" in new_risk_layer_fields:

                                    new_risk_layer.startEditing()

                                    for feature in new_risk_layer.getFeatures():
                                        try:
                                            if feature["r_indiv"]:
                                                feature["r_indiv"] = float(feature["r_indiv"])
                                                new_risk_layer.updateFeature(feature)
                                        except ValueError:
                                            print(f"Invalid value: {feature['r_indiv']}")
                                    new_risk_layer.commitChanges()

                                    ranges = []

                                    # Define first range
                                    symbol1 = QgsSymbol.defaultSymbol(new_risk_layer.geometryType())
                                    symbol1.setColor(QColor("lightgrey"))
                                    if key in ["ROAD", "RAILWAY", "CABLE_TRANSPORT", "LINE_PIPE"]:
                                        symbol1.setWidth(1.5)
                                    range1 = QgsRendererRange(0, 0.000001, symbol1, "< 1e-6")
                                    ranges.append(range1)

                                    # Define second range
                                    symbol2 = QgsSymbol.defaultSymbol(new_risk_layer.geometryType())
                                    if key in ["ROAD", "RAILWAY", "CABLE_TRANSPORT", "LINE_PIPE"]:
                                        symbol2.setWidth(1.5)
                                    symbol2.setColor(QColor("orange"))
                                    range2 = QgsRendererRange(0.000001, 0.00001, symbol2, "1e-6 - 1e-5")
                                    ranges.append(range2)

                                    # Define third range
                                    symbol3 = QgsSymbol.defaultSymbol(new_risk_layer.geometryType())
                                    if key in ["ROAD", "RAILWAY", "CABLE_TRANSPORT", "LINE_PIPE"]:
                                        symbol3.setWidth(1.5)
                                    symbol3.setColor(QColor("red"))
                                    range3 = QgsRendererRange(0.00001, 1, symbol3, "> 1e-5")
                                    ranges.append(range3)

                                    renderer = QgsGraduatedSymbolRenderer("r_indiv", ranges)

                                    # Set the renderer to the layer
                                    new_risk_layer.setRenderer(renderer)

                            # Add risk layer to map
                            QgsProject.instance().addMapLayer(new_risk_layer)
                            # USER INFO: SUCCESS
                            # StatusDialog(status_icon = self.scaled_pixmap_checkmark, message_title="Erfolg", message_text=f"\nRisiken erfolgreich visualisiert!\n").exec_()
                        else:
                            # USER INFO: Keine Szenen definiert
                            StatusDialog(status_icon = self.scaled_pixmap_warning, message_title="Fehler", message_text=f"\nKeine Szenen vorhanden!\n").exec_()
                else:
                    # USER INFO: Schadenpotential-Layer nicht synchronisiert
                    StatusDialog(status_icon = self.scaled_pixmap_warning, message_title="Fehler", message_text=f"\nBitte alle Schadenpotential-Layer synchronisieren oder entfernen!\n").exec_()
            else:
                # USER INFO: Kein Schadenpotential ausgewählt
                StatusDialog(status_icon = self.scaled_pixmap_warning, message_title="Fehler", message_text=f"\nBitte Schadenpotential-Layer auswählen, um Risiken zu visualisieren!\n").exec_()
        else:
            # USER INFO: Ungültige Anmeldedaten
            StatusDialog(status_icon = self.scaled_pixmap_warning, message_title="Fehler", message_text=f"\nUngültige Anmeldedaten! (Projektnummer oder API-Keys)\n").exec_()

    def on_dialog_closed(self):
        """Prevent that main dialog is opened more than 1x time."""
        self.main_dialog = None
        self.first_start = True

    def clean_up(self, file_path):
        """This method removes a file from the disk."""

        # Delete file
        if file_path is not None:
            if os.path.exists(file_path):
                os.remove(file_path)

    def run(self):
        """Run method that performs all the real work."""

        # ## PLUGIN LAUNCH
        # ------------------

        # Create the dialog with elements (after translation) and keep reference
        # Only create GUI ONCE in callback, so that it will only load when the plugin is started
        if self.first_start == True:
            self.first_start = False
            self.main_dialog = EconoMeDialog()
            self.main_dialog.finished.connect(self.on_dialog_closed)

        # Show the main plugin dialog
        self.main_dialog.show()
        self.main_dialog.BUTTON_PP_BUILDING.setFocusPolicy(Qt.NoFocus)
        self.main_dialog.BUTTON_ECONOME_PROJECT_CONNECT.setCursor(QCursor(Qt.PointingHandCursor))

        # Create a timer
        self.TIMER = QTimer()
        
        # create the pixmap
        path_to_checkmark = os.path.join(self.plugin_dir, "general-icons", "checkmark.png")
        pixmap_checkmark = QPixmap(path_to_checkmark)
        self.scaled_pixmap_checkmark = pixmap_checkmark.scaled(
            self.main_dialog.IMAGE_LABEL_SCENARIO_STATUS_FREE.size(),
            Qt.KeepAspectRatio,
            Qt.SmoothTransformation)
        path_to_warning = os.path.join(self.plugin_dir, "general-icons", "warning.png")
        pixmap_warning = QPixmap(path_to_warning)
        self.scaled_pixmap_warning = pixmap_warning.scaled(
            self.main_dialog.IMAGE_LABEL_SCENARIO_STATUS_FREE.size(),
            Qt.KeepAspectRatio,
            Qt.SmoothTransformation)

        # load the SELECTED_LAYERS_DICT
        try:
            project_folder = os.path.dirname(QgsProject.instance().fileName())
            json_file_path = os.path.join(project_folder, "SELECTED_LAYERS_DICT.json")
            with open(json_file_path, "r") as file:
                self.SELECTED_LAYERS_DICT = json.load(file)
        except:
            self.SELECTED_LAYERS_DICT = {"DAMAGE_POTENTIAL":{}, "INTENSITY_MAPS":{}}
            
        # ## ECONOME PROJECT LOGIN
        # -------------------------

        # Set project status
        self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.addItem("Nicht verbunden")
        # Get EconoMe user info (if existing)
        existing_user_info = self.get_econome_user_info()
        if existing_user_info:
            # Set user infos
            self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.setCurrentText(existing_user_info["ECONOME_PROJECT_NUMBER"])
            self.main_dialog.LINE_EDIT_ECONOME_USER_NAME.setText(existing_user_info["ECONOME_USER_NAME"])
            self.main_dialog.LINE_EDIT_ECONOME_PASSWORD.setText(existing_user_info["ECONOME_PASSWORD"])
            self.validate_econome_user_info()
        else:
            self.validate_econome_user_info()

        try:
            self.main_dialog.BUTTON_ECONOME_PROJECT_CONNECT.clicked.disconnect()
            self.main_dialog.LINE_EDIT_ECONOME_USER_NAME.textChanged.disconnect()
            self.main_dialog.LINE_EDIT_ECONOME_PASSWORD.textChanged.disconnect()
            self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.currentTextChanged.disconnect()
        except TypeError:
            print("No connections to disconnect.")

        self.main_dialog.BUTTON_ECONOME_PROJECT_CONNECT.clicked.connect(self.validate_econome_user_info)
        self.main_dialog.LINE_EDIT_ECONOME_USER_NAME.textChanged.connect(self.onChange_econome_user_info)
        self.main_dialog.LINE_EDIT_ECONOME_PASSWORD.textChanged.connect(self.onChange_econome_user_info)
        self.main_dialog.COMBO_BOX_ECONOME_PROJECT_LIST.currentTextChanged.connect(self.onChange_project_selection)

        # ## MAP LAYER COMBO BOXES
        # -------------------------

        # Create list of all map layer combo boxes
        map_layer_combo_boxes = [
            self.main_dialog.LAYER_DAM_POT_BUILDING,
            self.main_dialog.LAYER_DAM_POT_ROAD,
            self.main_dialog.LAYER_DAM_POT_RAILWAY,
            self.main_dialog.LAYER_DAM_POT_CABLE_TRANSPORT,
            self.main_dialog.LAYER_DAM_POT_LINE_PIPE,
            self.main_dialog.LAYER_DAM_POT_OTHER_ASSET,
            self.main_dialog.LAYER_INTENSITY_SCENARIO_FREE,
            self.main_dialog.LAYER_INTENSITY_SCENARIO_30,
            self.main_dialog.LAYER_INTENSITY_SCENARIO_100,
            self.main_dialog.LAYER_INTENSITY_SCENARIO_300,
            self.main_dialog.LAYER_INTENSITY_SCENARIO_EXTREME,
        ]

        # ## DAMAGE POTENTIAL
        # ---------------------

        # Disconnect any existing button connections (to avoid errors when plugin is re-opened)
        try:
            self.main_dialog.LAYER_DAM_POT_BUILDING.layerChanged.disconnect()
            self.main_dialog.LAYER_DAM_POT_ROAD.layerChanged.disconnect()
            self.main_dialog.LAYER_DAM_POT_RAILWAY.layerChanged.disconnect()
            self.main_dialog.LAYER_DAM_POT_CABLE_TRANSPORT.layerChanged.disconnect()
            self.main_dialog.LAYER_DAM_POT_LINE_PIPE.layerChanged.disconnect()
            self.main_dialog.LAYER_DAM_POT_OTHER_ASSET.layerChanged.disconnect()
            self.main_dialog.BUTTON_PP_BUILDING.clicked.disconnect()
            self.main_dialog.BUTTON_PP_ROAD.clicked.disconnect()
            self.main_dialog.BUTTON_PP_RAILWAY.clicked.disconnect()
            self.main_dialog.BUTTON_PP_CABLE_TRANSPORT.clicked.disconnect()
            self.main_dialog.BUTTON_PP_LINE_PIPE.clicked.disconnect()
            self.main_dialog.BUTTON_PP_OTHER_ASSET.clicked.disconnect()
            self.main_dialog.BUTTON_DAM_POT_SYNC_BUILDING.clicked.disconnect()
            self.main_dialog.BUTTON_DAM_POT_SYNC_ROAD.clicked.disconnect()
            self.main_dialog.BUTTON_DAM_POT_SYNC_RAILWAY.clicked.disconnect()
            self.main_dialog.BUTTON_DAM_POT_SYNC_CABLE_TRANSPORT.clicked.disconnect()
            self.main_dialog.BUTTON_DAM_POT_SYNC_LINE_PIPE.clicked.disconnect()
            self.main_dialog.BUTTON_DAM_POT_SYNC_OTHER_ASSET.clicked.disconnect()
            self.main_dialog.BUTTON_DAM_POT_DEL_BUILDING.clicked.disconnect()
            self.main_dialog.BUTTON_DAM_POT_DEL_ROAD.clicked.disconnect()
            self.main_dialog.BUTTON_DAM_POT_DEL_RAILWAY.clicked.disconnect()
            self.main_dialog.BUTTON_DAM_POT_DEL_CABLE_TRANSPORT.clicked.disconnect()
            self.main_dialog.BUTTON_DAM_POT_DEL_LINE_PIPE.clicked.disconnect()
            self.main_dialog.BUTTON_DAM_POT_DEL_OTHER_ASSET.clicked.disconnect()
        except TypeError:
            print("No connections to disconnect.")

        # Establish layer change events on damage potential layer dropdowns
        self.main_dialog.LAYER_DAM_POT_BUILDING.layerChanged.connect(self.update_asset_layer_status)
        self.main_dialog.LAYER_DAM_POT_ROAD.layerChanged.connect(self.update_asset_layer_status)
        self.main_dialog.LAYER_DAM_POT_RAILWAY.layerChanged.connect(self.update_asset_layer_status)
        self.main_dialog.LAYER_DAM_POT_CABLE_TRANSPORT.layerChanged.connect(self.update_asset_layer_status)
        self.main_dialog.LAYER_DAM_POT_LINE_PIPE.layerChanged.connect(self.update_asset_layer_status)
        self.main_dialog.LAYER_DAM_POT_OTHER_ASSET.layerChanged.connect(self.update_asset_layer_status)

        # Establish damage potential attribute profiles button connections & pointing cursor
        self.main_dialog.BUTTON_PP_BUILDING.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_PP_BUILDING.clicked.connect(
            lambda: self.set_parameter_profiles("building"))
        self.main_dialog.BUTTON_PP_ROAD.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_PP_ROAD.clicked.connect(
            lambda: self.set_parameter_profiles("road"))
        self.main_dialog.BUTTON_PP_RAILWAY.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_PP_RAILWAY.clicked.connect(
            lambda: self.set_parameter_profiles("railway"))
        self.main_dialog.BUTTON_PP_CABLE_TRANSPORT.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_PP_CABLE_TRANSPORT.clicked.connect(
            lambda: self.set_parameter_profiles("cable_transport"))
        self.main_dialog.BUTTON_PP_LINE_PIPE.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_PP_LINE_PIPE.clicked.connect(
            lambda: self.set_parameter_profiles("line_pipe"))
        self.main_dialog.BUTTON_PP_OTHER_ASSET.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_PP_OTHER_ASSET.clicked.connect(
            lambda: self.set_parameter_profiles("other_asset"))

        # Establish damage potential synchronization button connections & pointing cursor
        self.main_dialog.BUTTON_DAM_POT_SYNC_BUILDING.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_DAM_POT_SYNC_BUILDING.clicked.connect(lambda: self.synchronize_damage_potential(self.main_dialog.LAYER_DAM_POT_BUILDING.currentLayer(), "building", self.main_dialog.IMAGE_LABEL_DAM_POT_STATUS_BUILDING))
        self.main_dialog.BUTTON_DAM_POT_SYNC_ROAD.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_DAM_POT_SYNC_ROAD.clicked.connect(lambda: self.synchronize_damage_potential(self.main_dialog.LAYER_DAM_POT_ROAD.currentLayer(), "road", self.main_dialog.IMAGE_LABEL_DAM_POT_STATUS_ROAD))
        self.main_dialog.BUTTON_DAM_POT_SYNC_RAILWAY.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_DAM_POT_SYNC_RAILWAY.clicked.connect(lambda: self.synchronize_damage_potential(self.main_dialog.LAYER_DAM_POT_RAILWAY.currentLayer(), "railway", self.main_dialog.IMAGE_LABEL_DAM_POT_STATUS_RAILWAY))
        self.main_dialog.BUTTON_DAM_POT_SYNC_CABLE_TRANSPORT.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_DAM_POT_SYNC_CABLE_TRANSPORT.clicked.connect(lambda: self.synchronize_damage_potential(self.main_dialog.LAYER_DAM_POT_CABLE_TRANSPORT.currentLayer(), "cable_transport", self.main_dialog.IMAGE_LABEL_DAM_POT_STATUS_CABLE_TRANSPORT))
        self.main_dialog.BUTTON_DAM_POT_SYNC_LINE_PIPE.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_DAM_POT_SYNC_LINE_PIPE.clicked.connect(lambda: self.synchronize_damage_potential(self.main_dialog.LAYER_DAM_POT_LINE_PIPE.currentLayer(), "line_pipe", self.main_dialog.IMAGE_LABEL_DAM_POT_STATUS_LINE_PIPE))
        self.main_dialog.BUTTON_DAM_POT_SYNC_OTHER_ASSET.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_DAM_POT_SYNC_OTHER_ASSET.clicked.connect(lambda: self.synchronize_damage_potential(self.main_dialog.LAYER_DAM_POT_OTHER_ASSET.currentLayer(), "other_asset", self.main_dialog.IMAGE_LABEL_DAM_POT_STATUS_OTHER_ASSET))

        # Establish click connections on damage potential delete button connections & pointing cursor
        self.main_dialog.BUTTON_DAM_POT_DEL_BUILDING.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_DAM_POT_DEL_BUILDING.clicked.connect(
            lambda: self.delete_damage_potential_assets("building", self.main_dialog.IMAGE_LABEL_DAM_POT_STATUS_BUILDING))
        self.main_dialog.BUTTON_DAM_POT_DEL_ROAD.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_DAM_POT_DEL_ROAD.clicked.connect(
            lambda: self.delete_damage_potential_assets("road", self.main_dialog.IMAGE_LABEL_DAM_POT_STATUS_ROAD))
        self.main_dialog.BUTTON_DAM_POT_DEL_RAILWAY.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_DAM_POT_DEL_RAILWAY.clicked.connect(
            lambda: self.delete_damage_potential_assets("railway", self.main_dialog.IMAGE_LABEL_DAM_POT_STATUS_RAILWAY))
        self.main_dialog.BUTTON_DAM_POT_DEL_CABLE_TRANSPORT.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_DAM_POT_DEL_CABLE_TRANSPORT.clicked.connect(
            lambda: self.delete_damage_potential_assets("cable_transport", self.main_dialog.IMAGE_LABEL_DAM_POT_STATUS_CABLE_TRANSPORT))
        self.main_dialog.BUTTON_DAM_POT_DEL_LINE_PIPE.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_DAM_POT_DEL_LINE_PIPE.clicked.connect(
            lambda: self.delete_damage_potential_assets("line_pipe", self.main_dialog.IMAGE_LABEL_DAM_POT_STATUS_LINE_PIPE))
        self.main_dialog.BUTTON_DAM_POT_DEL_OTHER_ASSET.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_DAM_POT_DEL_OTHER_ASSET.clicked.connect(
            lambda: self.delete_damage_potential_assets("other_asset", self.main_dialog.IMAGE_LABEL_DAM_POT_STATUS_OTHER_ASSET))
        
        # Only allow vector layers and add an empty item as default to each map layer combo box
        self.main_dialog.LAYER_DAM_POT_BUILDING.setFilters(QgsMapLayerProxyModel.PolygonLayer | QgsMapLayerProxyModel.PointLayer)
        self.main_dialog.LAYER_DAM_POT_BUILDING.setAllowEmptyLayer(True)
        self.main_dialog.LAYER_DAM_POT_BUILDING.setCurrentIndex(0)
        self.main_dialog.LAYER_DAM_POT_ROAD.setFilters(QgsMapLayerProxyModel.LineLayer)
        self.main_dialog.LAYER_DAM_POT_ROAD.setAllowEmptyLayer(True)
        self.main_dialog.LAYER_DAM_POT_ROAD.setCurrentIndex(0)
        self.main_dialog.LAYER_DAM_POT_RAILWAY.setFilters(QgsMapLayerProxyModel.LineLayer)
        self.main_dialog.LAYER_DAM_POT_RAILWAY.setAllowEmptyLayer(True)
        self.main_dialog.LAYER_DAM_POT_RAILWAY.setCurrentIndex(0)
        self.main_dialog.LAYER_DAM_POT_CABLE_TRANSPORT.setFilters(QgsMapLayerProxyModel.LineLayer)
        self.main_dialog.LAYER_DAM_POT_CABLE_TRANSPORT.setAllowEmptyLayer(True)
        self.main_dialog.LAYER_DAM_POT_CABLE_TRANSPORT.setCurrentIndex(0)
        self.main_dialog.LAYER_DAM_POT_LINE_PIPE.setFilters(QgsMapLayerProxyModel.LineLayer)
        self.main_dialog.LAYER_DAM_POT_LINE_PIPE.setAllowEmptyLayer(True)
        self.main_dialog.LAYER_DAM_POT_LINE_PIPE.setCurrentIndex(0)
        self.main_dialog.LAYER_DAM_POT_OTHER_ASSET.setFilters(QgsMapLayerProxyModel.PolygonLayer)
        self.main_dialog.LAYER_DAM_POT_OTHER_ASSET.setAllowEmptyLayer(True)
        self.main_dialog.LAYER_DAM_POT_OTHER_ASSET.setCurrentIndex(0)
        
            
        # set saved layer names for damage potential
        for damage_potential_type, saved_layer_name in self.SELECTED_LAYERS_DICT["DAMAGE_POTENTIAL"].items():
            combo_box_layer_widget = getattr(self.main_dialog, f"LAYER_DAM_POT_{damage_potential_type}")
            try:
                found_layers = QgsProject.instance().mapLayersByName(saved_layer_name)
                if len(found_layers) == 1:
                    combo_box_layer_widget.setLayer(found_layers[0])
                else:
                    issue = f"Could not set layer: None or multiple layers found with the name {saved_layer_name}"
                    print(issue)
                    # TODO: error message popup to warn the user that he has multiple layers with the same name
            except:
                pass
            
        # ## SITUATION DROPDOWNS
        # ---------------------
        # Disconnect any existing dropdown events (to avoid errors when plugin is re-opened)
        try:
            self.main_dialog.COMBO_BOX_HAZARD_PROCESS.currentIndexChanged.disconnect()
            self.main_dialog.COMBO_BOX_MEASURE_VARIANT.currentIndexChanged.disconnect()
        except TypeError:
            print("No connections to disconnect.")
        
        # Establish dropdown events on situation dropdowns
        self.main_dialog.COMBO_BOX_HAZARD_PROCESS.currentIndexChanged.connect(self.update_situation_status)
        self.main_dialog.COMBO_BOX_MEASURE_VARIANT.currentIndexChanged.connect(self.update_situation_status)
        
        
        # ## INTENSITY LAYERS
        # ---------------------

        # Disconnect any existing layer change events (to avoid errors when plugin is re-opened)
        try:
            self.main_dialog.LAYER_INTENSITY_SCENARIO_FREE.layerChanged.disconnect()
            self.main_dialog.LAYER_INTENSITY_SCENARIO_30.layerChanged.disconnect()
            self.main_dialog.LAYER_INTENSITY_SCENARIO_100.layerChanged.disconnect()
            self.main_dialog.LAYER_INTENSITY_SCENARIO_300.layerChanged.disconnect()
            self.main_dialog.LAYER_INTENSITY_SCENARIO_EXTREME.layerChanged.disconnect()
        except TypeError:
            print("No connections to disconnect.")

        # clear the intensity layer combo boxes
        self.main_dialog.LAYER_INTENSITY_SCENARIO_FREE.setFilters(QgsMapLayerProxyModel.PolygonLayer)
        self.main_dialog.LAYER_INTENSITY_SCENARIO_FREE.setAllowEmptyLayer(True)
        self.main_dialog.LAYER_INTENSITY_SCENARIO_FREE.setCurrentIndex(0)
        self.main_dialog.LAYER_INTENSITY_SCENARIO_30.setFilters(QgsMapLayerProxyModel.PolygonLayer)
        self.main_dialog.LAYER_INTENSITY_SCENARIO_30.setAllowEmptyLayer(True)
        self.main_dialog.LAYER_INTENSITY_SCENARIO_30.setCurrentIndex(0)
        self.main_dialog.LAYER_INTENSITY_SCENARIO_100.setFilters(QgsMapLayerProxyModel.PolygonLayer)
        self.main_dialog.LAYER_INTENSITY_SCENARIO_100.setAllowEmptyLayer(True)
        self.main_dialog.LAYER_INTENSITY_SCENARIO_100.setCurrentIndex(0)
        self.main_dialog.LAYER_INTENSITY_SCENARIO_300.setFilters(QgsMapLayerProxyModel.PolygonLayer)
        self.main_dialog.LAYER_INTENSITY_SCENARIO_300.setAllowEmptyLayer(True)
        self.main_dialog.LAYER_INTENSITY_SCENARIO_300.setCurrentIndex(0)
        self.main_dialog.LAYER_INTENSITY_SCENARIO_EXTREME.setFilters(QgsMapLayerProxyModel.PolygonLayer)
        self.main_dialog.LAYER_INTENSITY_SCENARIO_EXTREME.setAllowEmptyLayer(True)
        self.main_dialog.LAYER_INTENSITY_SCENARIO_EXTREME.setCurrentIndex(0)
        
        # Establish layer change events on intensity layers dropdowns
        self.main_dialog.LAYER_INTENSITY_SCENARIO_FREE.layerChanged.connect(self.update_intensity_layer_status)
        self.main_dialog.LAYER_INTENSITY_SCENARIO_30.layerChanged.connect(self.update_intensity_layer_status)
        self.main_dialog.LAYER_INTENSITY_SCENARIO_100.layerChanged.connect(self.update_intensity_layer_status)
        self.main_dialog.LAYER_INTENSITY_SCENARIO_300.layerChanged.connect(self.update_intensity_layer_status)
        self.main_dialog.LAYER_INTENSITY_SCENARIO_EXTREME.layerChanged.connect(self.update_intensity_layer_status)
        
        # see which scenarios are active and activate/deactivate the widgets accordingly and set saved layer names for intensity maps
        self.update_situation_status()

        # Establish connections on intensity layers delete buttons
        self.main_dialog.BUTTON_DEL_AFFECTED_ASSETS_FREE.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_DEL_AFFECTED_ASSETS_FREE.clicked.connect(lambda: self.delete_damage_potential_affected_assets(self.main_dialog.LABEL_SCENARIO_FREE.text().strip('-jährlich')))
        self.main_dialog.BUTTON_DEL_AFFECTED_ASSETS_30.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_DEL_AFFECTED_ASSETS_30.clicked.connect(lambda: self.delete_damage_potential_affected_assets("30"))
        self.main_dialog.BUTTON_DEL_AFFECTED_ASSETS_100.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_DEL_AFFECTED_ASSETS_100.clicked.connect(lambda: self.delete_damage_potential_affected_assets("100"))
        self.main_dialog.BUTTON_DEL_AFFECTED_ASSETS_300.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_DEL_AFFECTED_ASSETS_300.clicked.connect(lambda: self.delete_damage_potential_affected_assets("300"))
        self.main_dialog.BUTTON_DEL_AFFECTED_ASSETS_EXTREME.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_DEL_AFFECTED_ASSETS_EXTREME.clicked.connect(lambda: self.delete_damage_potential_affected_assets(self.main_dialog.LABEL_SCENARIO_EXTREME.text().strip('-jährlich')))
        
        # ## RISK CALCULATION AND VISUALIZATION
        # --------------------------------------

        # Disconnect any existing button connections (to avoid errors when plugin is re-opened)
        try:
            self.main_dialog.BUTTON_UPLOAD_CONSEQUENCE_ANALYSIS.clicked.disconnect()
            self.main_dialog.BUTTON_VISUALIZE_RISKS.clicked.disconnect()
        except TypeError:
            print("No connections to disconnect.")

        # Establish risk calculation and visualization button connections & pointing cursor
        self.main_dialog.BUTTON_UPLOAD_CONSEQUENCE_ANALYSIS.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_UPLOAD_CONSEQUENCE_ANALYSIS.clicked.connect(lambda: self.upload_consequence_analysis())
        self.main_dialog.BUTTON_VISUALIZE_RISKS.setCursor(QCursor(Qt.PointingHandCursor))
        self.main_dialog.BUTTON_VISUALIZE_RISKS.clicked.connect(lambda: self.visualize_risks())

        # ## QGIS EXIT
        # --------------
        try:
            QCoreApplication.instance().aboutToQuit.disconnect()
        except TypeError:
            print("No connections to disconnect.")

        QCoreApplication.instance().aboutToQuit.connect(lambda: self.clean_up(self.ECONOME_USER_INFO_FILE_PATH))

        # Run the main plugin dialog event loop
        result = self.main_dialog.exec_()
        # save the layer dict as json
        try:
            project_folder = os.path.dirname(QgsProject.instance().fileName())
            json_file_path = os.path.join(project_folder, "SELECTED_LAYERS_DICT.json")
            with open(json_file_path, 'w') as json_file:
                json.dump(self.SELECTED_LAYERS_DICT, json_file)
            print("Saved SELECTED_LAYERS_DICT.json")
        except:
            print("Could not save SELECTED_LAYERS_DICT.json")
            
    def update_asset_layer_status(self):
        # Get sender info
        sender_widget = self.main_dialog.sender()
        dmage_potential_type = sender_widget.objectName().split("DAM_POT_")[-1]
        
        # clear image label
        image_label_widget = getattr(self.main_dialog, f"IMAGE_LABEL_DAM_POT_STATUS_{dmage_potential_type}")
        image_label_widget.clear()
        
        # save selected layer name
        current_layer_name = self.main_dialog.sender().currentText()
        self.SELECTED_LAYERS_DICT['DAMAGE_POTENTIAL'].update({dmage_potential_type: current_layer_name})
        
    def update_intensity_layer_status(self):
        # Get sender info
        sender_widget = self.main_dialog.sender()
        annuality = sender_widget.objectName().split("INTENSITY_SCENARIO_")[-1]
        
        # fill dropdown with layer fields
        current_layer = self.main_dialog.sender().currentLayer()
        combo_box_name = f"COMBO_BOX_INTENSITY_COLUMN_NAME_{annuality}"
        combo_box_widget = getattr(self.main_dialog, combo_box_name)
        if current_layer:
            combo_box_widget.clear()
            combo_box_widget.addItem("")
            # Add field names to dropdown
            for field in current_layer.fields():
                combo_box_widget.addItem(field.name())
                combo_box_widget.setCurrentIndex(0)
        else:
            combo_box_widget.clear()
            
        # save selected layer name
        current_layer_name = self.main_dialog.sender().currentText()
        current_process_id = self.main_dialog.COMBO_BOX_HAZARD_PROCESS.currentText().split(", ")[0]
        current_variant_id = self.main_dialog.COMBO_BOX_MEASURE_VARIANT.currentText().split(", ")[0]
        if current_variant_id != "" and current_process_id != "":
            key = f"{current_process_id}_{current_variant_id}_{annuality}"
            self.SELECTED_LAYERS_DICT['INTENSITY_MAPS'].update({key: current_layer_name})
    
    def update_situation_status(self):
        # get situation info
        current_process_id = self.main_dialog.COMBO_BOX_HAZARD_PROCESS.currentText().split(", ")[0]
        current_variant_id = self.main_dialog.COMBO_BOX_MEASURE_VARIANT.currentText().split(", ")[0]
        
        # get process info
        get_process = requests.get(self.ECONOME_API_PROJECT_BASE_PATH + f"/{self.ECONOME_PROJECT_NUMBER}" + "/process" + f"/{current_process_id}", auth=HTTPBasicAuth(str(self.ECONOME_USER_NAME), str(self.ECONOME_PASSWORD)))
        if len(json.loads(get_process.text)) != 0:
            SCENARIOS = {key : value["Active"] for key, value in json.loads(get_process.text).items() if key.startswith('Annuality') and value is not None}
            for annuality_key, active in SCENARIOS.items():
                annuality_str = annuality_key.strip("Annuality")
                annuality = annuality_str.upper() if annuality_str in ["Extreme", "Free"] else int(annuality_str)
                # get widgets
                layer_combo_box_widget = getattr(self.main_dialog, f"LAYER_INTENSITY_SCENARIO_{annuality}")
                column_name_combo_box_widget = getattr(self.main_dialog, f"COMBO_BOX_INTENSITY_COLUMN_NAME_{annuality}")
                delete_button_widget = getattr(self.main_dialog, f"BUTTON_DEL_AFFECTED_ASSETS_{annuality}")
                checkmark_widget = getattr(self.main_dialog, f"IMAGE_LABEL_SCENARIO_STATUS_{annuality}")
                
                # deselect the layer
                layer_combo_box_widget.setCurrentIndex(0)
                
                # if scenario is active, we activate the widgets, add the checkmark and fill the dropdown with the saved layer
                if active:
                    layer_combo_box_widget.setEnabled(True)
                    column_name_combo_box_widget.setEnabled(True)
                    delete_button_widget.setEnabled(True)
                    checkmark_widget.setPixmap(self.scaled_pixmap_checkmark)
                
                    # auto-select the saved layer
                    key = f"{current_process_id}_{current_variant_id}_{annuality}"
                    saved_layer_name = self.SELECTED_LAYERS_DICT['INTENSITY_MAPS'].get(key)
                    try:
                        found_layers = QgsProject.instance().mapLayersByName(saved_layer_name)
                        if len(found_layers) == 1:
                            layer_combo_box_widget.setLayer(found_layers[0])
                        else:
                            issue = f"Could not set layer: None or multiple layers found with the name {saved_layer_name}"
                            print(issue)
                            # TODO: error message popup to warn the user that he has multiple layers with the same name
                    except:
                        pass
                
                # if scenario is inactive, we deactivate all widgets of this scenario and remove the checkmark
                else:
                    layer_combo_box_widget.setEnabled(False)
                    column_name_combo_box_widget.setEnabled(False)
                    delete_button_widget.setEnabled(False)
                    checkmark_widget.clear()
